# -*- coding: utf-8 -*-

from __future__ import annotations

__all__ = ["kalman_filter"]

from typing import Any, Optional, Tuple

import jax
import jax.numpy as jnp
import numpy as np

from tinygp.helpers import JAXArray, dataclass
from tinygp.kernels.base import Kernel
from tinygp.noise import Diagonal, Noise
from tinygp.solvers.solver import Solver


@dataclass
class KalmanSolver(Solver):
    """A scalable solver that uses Kalman filtering

    This implementation is very limited and it is meant primarily

    You generally won't instantiate this object directly but, if you do, you'll
    probably want to use the :func:`KalmanSolver.init` method instead of the
    usual constructor.
    """

    X: JAXArray
    A: JAXArray
    H: JAXArray
    s: JAXArray
    K: JAXArray

    @classmethod
    def init(
        cls,
        kernel: Kernel,
        X: JAXArray,
        noise: Noise,
        *,
        covariance: Optional[Any] = None,
    ) -> "KalmanSolver":
        """Build a :class:`KalmanSolver` for a given kernel and coordinates

        Args:
            kernel: The kernel function. This must be an instance of a subclass
                of :class:`tinygp.kernels.quasisep.Quasisep`.
            X: The input coordinates.
            noise: The noise model for the process. This must be diagonal for
                this solver.
            covariance: Not yet supported by this solver.
        """
        from tinygp.kernels.quasisep import Quasisep

        assert isinstance(kernel, Quasisep)
        assert isinstance(noise, Diagonal)
        assert covariance is None

        Pinf = kernel.stationary_covariance()
        A = jax.vmap(kernel.transition_matrix)(
            jax.tree_util.tree_map(lambda y: jnp.append(y[0], y[:-1]), X), X
        )
        H = jax.vmap(kernel.observation_model)(X)
        s, K = kalman_gains(Pinf, A, H, noise.diag)
        return cls(X=X, A=A, H=H, s=s, K=K)

    def variance(self) -> JAXArray:
        raise NotImplementedError

    def covariance(self) -> JAXArray:
        raise NotImplementedError

    def normalization(self) -> JAXArray:
        return 0.5 * jnp.sum(jnp.log(2 * np.pi * self.s))

    def solve_triangular(
        self, y: JAXArray, *, transpose: bool = False
    ) -> JAXArray:
        assert not transpose
        return kalman_filter(self.A, self.H, self.K, y) / jnp.sqrt(self.s)

    def dot_triangular(self, y: JAXArray) -> JAXArray:
        raise NotImplementedError

    def condition(
        self, kernel: Kernel, X_test: Optional[JAXArray], noise: Noise
    ) -> Any:
        raise NotImplementedError


@jax.jit
def kalman_gains(
    Pinf: JAXArray, A: JAXArray, H: JAXArray, diag: JAXArray
) -> Tuple[JAXArray, JAXArray]:
    def step(carry, data):  # type: ignore
        Pp = carry
        Ak, hk, dk = data

        Pn = Pinf + Ak.transpose() @ (Pp - Pinf) @ Ak
        tmp = Pn @ hk
        sk = hk @ tmp + dk
        Kk = tmp / sk
        Pk = Pn - sk * jnp.outer(Kk, Kk)

        return Pk, (sk, Kk)

    init = Pinf
    return jax.lax.scan(step, init, (A, H, diag))[1]


@jax.jit
def kalman_filter(
    A: JAXArray, H: JAXArray, K: JAXArray, y: JAXArray
) -> JAXArray:
    def step(carry, data):  # type: ignore
        mp = carry
        Ak, hk, Kk, yk = data

        mn = Ak.transpose() @ mp
        vk = yk - hk @ mn
        mk = mn + Kk * vk

        return mk, vk

    init = jnp.zeros_like(H[0])
    return jax.lax.scan(step, init, (A, H, K, y))[1]



# -*- coding: utf-8 -*-

from __future__ import annotations

__all__ = [
    "Kernel",
    "Conditioned",
    "Custom",
    "Sum",
    "Product",
    "Constant",
    "DotProduct",
    "Polynomial",
]

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Optional, Sequence, Union

import jax
import jax.numpy as jnp

from tinygp.helpers import JAXArray, dataclass, field

if TYPE_CHECKING:
    from tinygp.solvers.solver import Solver


Axis = Union[int, Sequence[int]]


class Kernel(metaclass=ABCMeta):
    """The base class for all kernel implementations

    This subclass provides default implementations to add and multiply kernels.
    Subclasses should accept parameters in their ``__init__`` and then override
    :func:`Kernel.evaluate` with custom behavior.
    """

    if TYPE_CHECKING:

        def __init__(self, *args: Any, **kwargs: Any) -> None:
            pass

    @abstractmethod
    def evaluate(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """Evaluate the kernel at a pair of input coordinates

        This should be overridden be subclasses to return the kernel-specific
        value. Two things to note:

        1. Users shouldn't generally call :func:`Kernel.evaluate`. Instead,
           always "call" the kernel instance directly; for example, you can
           evaluate the Matern-3/2 kernel using ``Matern32(1.5)(x1, x2)``, for
           arrays of input coordinates ``x1`` and ``x2``.
        2. When implementing a custom kernel, this method should treat ``X1``
           and ``X2`` as single datapoints. In other words, these inputs will
           typically either be scalars of have shape ``n_dim``, where ``n_dim``
           is the number of input dimensions, rather than ``n_data`` or
           ``(n_data, n_dim)``, and you should let the :class:`Kernel` ``vmap``
           magic handle all the broadcasting for you.
        """
        raise NotImplementedError

    def evaluate_diag(self, X: JAXArray) -> JAXArray:
        """Evaluate the kernel on its diagonal

        The default implementation simply calls :func:`Kernel.evaluate` with
        ``X`` as both arguments, but subclasses can use this to make diagonal
        calcuations more efficient.
        """
        return self.evaluate(X, X)

    def matmul(
        self,
        X1: JAXArray,
        X2: Optional[JAXArray] = None,
        y: Optional[JAXArray] = None,
    ) -> JAXArray:
        if y is None:
            assert X2 is not None
            y = X2
            X2 = None

        if X2 is None:
            X2 = X1

        return jnp.dot(self(X1, X2), y)

    def __call__(
        self, X1: JAXArray, X2: Optional[JAXArray] = None
    ) -> JAXArray:
        if X2 is None:
            k = jax.vmap(self.evaluate_diag, in_axes=0)(X1)
            if k.ndim != 1:
                raise ValueError(
                    "Invalid kernel diagonal shape: "
                    f"expected ndim = 1, got ndim={k.ndim} "
                    "check the dimensions of parameters and custom kernels"
                )
            return k
        k = jax.vmap(
            jax.vmap(self.evaluate, in_axes=(None, 0)), in_axes=(0, None)
        )(X1, X2)
        if k.ndim != 2:
            raise ValueError(
                "Invalid kernel shape: "
                f"expected ndim = 2, got ndim={k.ndim} "
                "check the dimensions of parameters and custom kernels"
            )
        return k

    def __add__(self, other: Union["Kernel", JAXArray]) -> "Kernel":
        if isinstance(other, Kernel):
            return Sum(self, other)
        return Sum(self, Constant(other))

    def __radd__(self, other: Any) -> "Kernel":
        # We'll hit this first branch when using the `sum` function
        if other == 0:
            return self
        if isinstance(other, Kernel):
            return Sum(other, self)
        return Sum(Constant(other), self)

    def __mul__(self, other: Union["Kernel", JAXArray]) -> "Kernel":
        if isinstance(other, Kernel):
            return Product(self, other)
        return Product(self, Constant(other))

    def __rmul__(self, other: Any) -> "Kernel":
        if isinstance(other, Kernel):
            return Product(other, self)
        return Product(Constant(other), self)


@dataclass
class Conditioned(Kernel):
    """A kernel used when conditioning a process on data

    Args:
        X: The coordinates of the data.
        scale_tril: The lower Cholesky factor of the base process' kernel
            matrix.
        kernel: The predictive kerenl; this will generally be the kernel from
            the kernel used by the original process.
    """

    X: JAXArray
    solver: Solver
    kernel: Kernel

    def evaluate(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        kernel_vec = jax.vmap(self.kernel.evaluate, in_axes=(0, None))
        K1 = self.solver.solve_triangular(kernel_vec(self.X, X1))
        K2 = self.solver.solve_triangular(kernel_vec(self.X, X2))
        return self.kernel.evaluate(X1, X2) - K1.transpose() @ K2

    def evaluate_diag(self, X: JAXArray) -> JAXArray:
        kernel_vec = jax.vmap(self.kernel.evaluate, in_axes=(0, None))
        K = self.solver.solve_triangular(kernel_vec(self.X, X))
        return self.kernel.evaluate_diag(X) - K.transpose() @ K


@dataclass
class Custom(Kernel):
    """A custom kernel class implemented as a callable

    Args:
        function: A callable with a signature and behavior that matches
            :func:`Kernel.evaluate`.
    """

    function: Callable[[Any, Any], Any]

    def evaluate(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        return self.function(X1, X2)


@dataclass
class Sum(Kernel):
    """A helper to represent the sum of two kernels"""

    kernel1: Kernel
    kernel2: Kernel

    def evaluate(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        return self.kernel1.evaluate(X1, X2) + self.kernel2.evaluate(X1, X2)


@dataclass
class Product(Kernel):
    """A helper to represent the product of two kernels"""

    kernel1: Kernel
    kernel2: Kernel

    def evaluate(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        return self.kernel1.evaluate(X1, X2) * self.kernel2.evaluate(X1, X2)


@dataclass
class Constant(Kernel):
    r"""This kernel returns the constant

    .. math::

        k(\mathbf{x}_i,\,\mathbf{x}_j) = c

    where :math:`c` is a parameter.

    Args:
        c: The parameter :math:`c` in the above equation.
    """

    value: JAXArray

    def __post_init__(self) -> None:
        if jnp.ndim(self.value) != 0:
            raise ValueError("The value of a constant kernel must be a scalar")

    def evaluate(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        return self.value


@dataclass
class DotProduct(Kernel):
    r"""The dot product kernel

    .. math::

        k(\mathbf{x}_i,\,\mathbf{x}_j) = \mathbf{x}_i \cdot \mathbf{x}_j

    with no parameters.
    """

    def evaluate(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        if jnp.ndim(X1) == 0:
            return X1 * X2
        return X1 @ X2


@dataclass
class Polynomial(Kernel):
    r"""A polynomial kernel

    .. math::

        k(\mathbf{x}_i,\,\mathbf{x}_j) = [(\mathbf{x}_i / \ell) \cdot
            (\mathbf{x}_j / \ell) + \sigma^2]^P

    Args:
        order: The power :math:`P`.
        scale: The parameter :math:`\ell`.
        sigma: The parameter :math:`\sigma`.
    """

    order: JAXArray
    scale: JAXArray = field(default_factory=lambda: jnp.ones(()))
    sigma: JAXArray = field(default_factory=lambda: jnp.zeros(()))

    def evaluate(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        return (
            (X1 / self.scale) @ (X2 / self.scale) + jnp.square(self.sigma)
        ) ** self.order



# -*- coding: utf-8 -*-
# mypy: ignore-errors

import jax
import jax.numpy as jnp
import numpy as np

from tinygp.kernels import quasisep


def test_matmul():
    random = np.random.default_rng(1234)
    x1 = np.sort(random.uniform(0, 10, 100))
    x2 = np.sort(random.uniform(2, 8, 75))
    kernel = quasisep.Matern52(sigma=1.5, scale=3.4)

    for (x1, x2) in [(x1, x2), (x1, x1), (x2, x1)]:
        y = np.sin(x2)[:, None]
        K = kernel(x1, x2)
        mat = kernel.to_general_qsm(x1, x2)
        np.testing.assert_allclose(mat.matmul(y), K @ y)



# -*- coding: utf-8 -*-
"""
In ``tinygp``, "transforms" are a powerful and relatively safe way to build
extremely expressive kernels without resorting to writing a fully fledged custom
kernel. More details can be found in the :ref:`transforms` tutorial.
"""

from __future__ import annotations

__all__ = ["Transform", "Linear", "Cholesky", "Subspace"]

from functools import partial
from typing import Any, Callable, Sequence, Union

import jax.numpy as jnp
from jax.scipy import linalg

from tinygp.helpers import JAXArray, dataclass
from tinygp.kernels.base import Kernel


@dataclass
class Transform(Kernel):
    """Apply a transformation to the input coordinates of the kernel

    Args:
        transform: (Callable): A callable object that accepts coordinates as
            inputs and returns transformed coordinates.
        kernel (Kernel): The kernel to use in the transformed space.
    """

    transform: Callable[[Any], Any]
    kernel: Kernel

    def evaluate(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        return self.kernel.evaluate(self.transform(X1), self.transform(X2))


@dataclass
class Linear(Kernel):
    """Apply a linear transformation to the input coordinates of the kernel

    For example, the following transformed kernels are all equivalent, but the
    second supports more flexible transformations:

    .. code-block:: python

        >>> import numpy as np
        >>> from tinygp import kernels, transforms
        >>> kernel0 = kernels.Matern32(4.5)
        >>> kernel1 = transforms.Linear(1.0 / 4.5, kernels.Matern32())
        >>> np.testing.assert_allclose(
        ...     kernel0.evaluate(0.5, 0.1), kernel1.evaluate(0.5, 0.1)
        ... )

    Args:
        scale (JAXArray): A 0-, 1-, or 2-dimensional array specifying the
            scale of this transform.
        kernel (Kernel): The kernel to use in the transformed space.
    """

    scale: JAXArray
    kernel: Kernel

    def evaluate(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        if jnp.ndim(self.scale) < 2:
            transform = partial(jnp.multiply, self.scale)
        elif jnp.ndim(self.scale) == 2:
            transform = partial(jnp.dot, self.scale)
        else:
            raise ValueError("'scale' must be 0-, 1-, or 2-dimensional")
        return self.kernel.evaluate(transform(X1), transform(X2))


@dataclass
class Cholesky(Kernel):
    """Apply a Cholesky transformation to the input coordinates of the kernel

    For example, the following transformed kernels are all equivalent, but the
    second supports more flexible transformations:

    .. code-block:: python

        >>> import numpy as np
        >>> from tinygp import kernels, transforms
        >>> kernel0 = kernels.Matern32(4.5)
        >>> kernel1 = transforms.Cholesky(4.5, kernels.Matern32())
        >>> np.testing.assert_allclose(
        ...     kernel0.evaluate(0.5, 0.1), kernel1.evaluate(0.5, 0.1)
        ... )

    Args:
        factor (JAXArray): A 0-, 1-, or 2-dimensional array specifying the
            Cholesky factor. If 2-dimensional, this must be a lower
            triangular matrix, but this is not checked.
        kernel (Kernel): The kernel to use in the transformed space.
    """

    factor: JAXArray
    kernel: Kernel

    def evaluate(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        if jnp.ndim(self.factor) < 2:
            transform = partial(jnp.multiply, 1.0 / self.factor)
        elif jnp.ndim(self.factor) == 2:
            transform = partial(
                linalg.solve_triangular, self.factor, lower=True
            )
        else:
            raise ValueError("'scale' must be 0-, 1-, or 2-dimensional")
        return self.kernel.evaluate(transform(X1), transform(X2))

    @classmethod
    def from_parameters(
        cls, diagonal: JAXArray, off_diagonal: JAXArray, kernel: Kernel
    ) -> "Cholesky":
        """Build a Cholesky transform with a sensible parameterization

        Args:
            diagonal (JAXArray): An ``(ndim,)`` array with the diagonal
                elements of ``factor``. These must be positive, but this
                is not checked.
            off_diagonal (JAXArray): An ``((ndim - 1) * ndim,)`` array
                with the off-diagonal elements of ``factor``.
            kernel (Kernel): The kernel to use in the transformed space.
        """
        ndim = diagonal.size
        if off_diagonal.size != ((ndim - 1) * ndim) // 2:
            raise ValueError(
                "Dimension mismatch: expected "
                f"(ndim-1)*ndim/2 = {((ndim - 1) * ndim) // 2} elements in "
                f"'off_diagonal'; got {off_diagonal.size}"
            )
        factor = jnp.zeros((ndim, ndim))
        factor = factor.at[jnp.diag_indices(ndim)].add(diagonal)
        factor = factor.at[jnp.tril_indices(ndim, -1)].add(off_diagonal)
        return cls(factor, kernel)


@dataclass
class Subspace(Kernel):
    """A kernel transform that selects a subset of the input dimensions

    For example, the following kernel only depends on the coordinates in the
    second (`1`-th) dimension:

    .. code-block:: python

        >>> import numpy as np
        >>> from tinygp import kernels, transforms
        >>> kernel = transforms.Subspace(1, kernels.Matern32())
        >>> np.testing.assert_allclose(
        ...     kernel.evaluate(np.array([0.5, 0.1]), np.array([-0.4, 0.7])),
        ...     kernel.evaluate(np.array([100.5, 0.1]), np.array([-70.4, 0.7])),
        ... )

    Args:
        axis: (Axis, optional): An integer or tuple of integers specifying the
            axes to select.
        kernel (Kernel): The kernel to use in the transformed space.
    """

    axis: Union[Sequence[int], int]
    kernel: Kernel

    def evaluate(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        return self.kernel.evaluate(X1[self.axis], X2[self.axis])



# -*- coding: utf-8 -*-
"""
In ``tinygp``, "solvers" provide a swappable low-level interface for
implementing the linear algebra required to execute Gaussian Process models. At
the moment, ``tinygp`` includes two solvers, but new solvers can be implemented
as external packages or as pull requests to the main ``tinygp`` GitHub project.
The two built in solvers are:

1. :class:`DirectSolver`: A solver that uses a naive approach to solving the
   required linear systems. This can still be pretty fast if you have access to
   a GPU that can run ``jax`` code. This is the default solver, and it can be
   used with any kernel implemented by ``tinygp``. Up to numerical precision,
   this is an *exact* solver.

2. :class:`QuasisepSolver`: An experiemental scalable solver that exploits the
   "quasiseparable" structure found in many GP covariance matrices to make the
   required linear algabra possible in linear scaling with the size of the
   dataset. These methods were previously implemented as part of the `celerite
   <https://celerite.readthedocs.io>`_ project.

Users generally won't instantiate these solvers directly and ``tinygp`` should
generally be able to figure out the best one to use. But you can use a specific
solver using the ``solver`` argument to :class:`tinygp.GaussianProcess` as
follows:

.. code-block:: python

    gp = tinygp.GaussianProcess(..., solver=tinygp.solvers.DirectSolver)

The details for the included solvers are given below, but this is a pretty
low-level feature and the details are definitely subject to change!
"""

__all__ = ["DirectSolver", "QuasisepSolver"]

from tinygp.solvers.direct import DirectSolver
from tinygp.solvers.quasisep import QuasisepSolver



# -*- coding: utf-8 -*-
# mypy: ignore-errors

import jax
import jax.numpy as jnp
import numpy as np
import pytest

from tinygp import GaussianProcess
from tinygp.kernels import quasisep
from tinygp.solvers import QuasisepSolver
from tinygp.solvers.kalman import KalmanSolver, kalman_filter, kalman_gains


@pytest.fixture
def random():
    return np.random.default_rng(84930)


@pytest.fixture
def data(random):
    x = np.sort(random.uniform(-3, 3, 50))
    y = np.sin(x)
    return x, y


@pytest.fixture(
    params=[
        quasisep.Matern32(sigma=1.8, scale=1.5),
        1.8**2 * quasisep.Matern32(1.5),
        quasisep.Matern52(sigma=1.8, scale=1.5),
        quasisep.Exp(sigma=1.8, scale=1.5),
        quasisep.Cosine(sigma=1.8, scale=1.5),
        quasisep.SHO(sigma=1.8, omega=1.5, quality=3.0),
        quasisep.SHO(sigma=1.8, omega=1.5, quality=0.2),
        quasisep.Celerite(1.1, 0.8, 0.9, 0.1),
        1.5 * quasisep.Matern52(1.5) + 0.3 * quasisep.Exp(1.5),
        quasisep.Matern52(1.5) * quasisep.SHO(omega=1.5, quality=0.1),
        1.5 * quasisep.Matern52(1.5) * quasisep.Celerite(1.1, 0.8, 0.9, 0.1),
        quasisep.CARMA.init(
            alpha=np.array([1.4, 2.3, 1.5]), beta=np.array([0.1, 0.5])
        ),
    ]
)
def kernel(request):
    return request.param


def test_filter(kernel, data):
    x, y = data
    diag = jnp.full_like(x, 0.1)

    logp0 = GaussianProcess(kernel, x, diag=diag).log_probability(y)

    Pinf = kernel.stationary_covariance()
    A = jax.vmap(kernel.transition_matrix)(jnp.append(x[0], x[:-1]), x)
    H = jax.vmap(kernel.observation_model)(x)
    s, K = kalman_gains(Pinf, A, H, diag)
    v = kalman_filter(A, H, K, y)
    logp = -0.5 * jnp.sum(jnp.square(v) / s + jnp.log(2 * jnp.pi * s))

    np.testing.assert_allclose(logp, logp0)


def test_consistent_with_direct(kernel, data):
    x, y = data
    gp1 = GaussianProcess(kernel, x, diag=0.1, solver=KalmanSolver)
    gp2 = GaussianProcess(kernel, x, diag=0.1, solver=QuasisepSolver)

    np.testing.assert_allclose(gp1.log_probability(y), gp2.log_probability(y))
    np.testing.assert_allclose(
        gp1.solver.normalization(), gp2.solver.normalization()
    )



# -*- coding: utf-8 -*-
# mypy: ignore-errors

import jax
import jax.numpy as jnp
import numpy as np
import pytest

from tinygp import kernels, noise
from tinygp.solvers import DirectSolver


@pytest.fixture
def random():
    return np.random.default_rng(1058390)


@pytest.fixture
def data(random):
    x1 = random.uniform(-3, 3, (50, 5))
    x2 = random.uniform(-5, 5, (50, 5))
    return x1, x2


def test_constant(data):
    x1, x2 = data

    # Check for dimension issues, either when instantiated...
    with pytest.raises(ValueError):
        kernels.Constant(jnp.ones(3))

    # ... or when multiplied
    with pytest.raises(ValueError):
        jnp.ones(3) * kernels.Matern32(1.5)

    # Check that multiplication has the expected behavior
    factor = 2.5
    k1 = kernels.Matern32(2.5)
    np.testing.assert_allclose(factor * k1(x1, x2), (factor * k1)(x1, x2))


def test_custom(data):
    x1, x2 = data

    # Check that known kernels work as expected
    scale = 1.5
    k1 = kernels.Custom(
        lambda X1, X2: jnp.exp(-0.5 * jnp.sum(jnp.square((X1 - X2) / scale)))
    )
    k2 = kernels.ExpSquared(scale)
    np.testing.assert_allclose(k1(x1, x2), k2(x1, x2))

    # Check that an invalid kernel raises as expected
    kernel = kernels.Custom(
        lambda X1, X2: jnp.exp(-0.5 * jnp.square((X1 - X2) / scale))
    )
    with pytest.raises(ValueError):
        kernel(x1, x2)


def test_ops(data):
    x1, x2 = data

    k1 = 1.5 * kernels.Matern32(2.5)
    k2 = 0.9 * kernels.ExpSineSquared(scale=1.5, gamma=0.3)

    np.testing.assert_allclose(k1(x1, x2) + k2(x1, x2), (k1 + k2)(x1, x2))
    np.testing.assert_allclose(k1(x1, x2) * k2(x1, x2), (k1 * k2)(x1, x2))


def test_conditioned(data):
    x1, x2 = data
    with jax.experimental.enable_x64():
        k1 = 1.5 * kernels.Matern32(2.5)
        k2 = 0.9 * kernels.ExpSineSquared(scale=1.5, gamma=0.3)
        K = k1(x1, x1) + 0.1 * jnp.eye(x1.shape[0])
        solver = DirectSolver.init(
            k1, x1, noise.Diagonal(jnp.full(x1.shape[0], 0.1))
        )
        cond = kernels.Conditioned(x1, solver, k2)
        np.testing.assert_allclose(
            cond(x1, x2),
            k2(x1, x2) - k2(x1, x1) @ jnp.linalg.solve(K, k2(x1, x2)),
        )


def test_dot_product(data):
    x1, x2 = data
    kernel = kernels.DotProduct()
    np.testing.assert_allclose(kernel(x1, x2), jnp.dot(x1, x2.T))
    np.testing.assert_allclose(
        kernel(x1[:, 0], x2[:, 0]), x1[:, 0][:, None] * x2[:, 0][None]
    )



# -*- coding: utf-8 -*-
# mypy: ignore-errors

from __future__ import annotations

__all__ = ["TinyDistribution"]

import jax.numpy as jnp
from numpyro.distributions import constraints
from numpyro.distributions.distribution import Distribution
from numpyro.distributions.util import (
    is_prng_key,
    lazy_property,
    validate_sample,
)


class TinyDistribution(Distribution):
    """Blah"""

    support = constraints.real_vector

    def __init__(self, gp, validate_args=None):
        self.gp = gp
        self.loc = gp.loc
        batch_shape = ()
        event_shape = jnp.shape(self.loc)
        super().__init__(
            batch_shape=batch_shape,
            event_shape=event_shape,
            validate_args=validate_args,
        )

    def sample(self, key, sample_shape=()):
        assert is_prng_key(key)
        return self.gp.sample(key, shape=sample_shape)

    @validate_sample
    def log_prob(self, value):
        return self.gp.log_probability(value)

    @lazy_property
    def covariance_matrix(self):
        return self.gp.covariance

    @lazy_property
    def precision_matrix(self):
        return self.gp.solver.solve_triangular(
            self.gp.solver.solve_triangular(jnp.eye(self.loc.shape[0])),
            transpose=True,
        )

    @property
    def mean(self):
        return self.loc

    @property
    def variance(self):
        return self.gp.variance

    def tree_flatten(self):
        return self.gp, None

    @classmethod
    def tree_unflatten(cls, aux_data, params):
        return cls(params)

    @staticmethod
    def infer_shapes(*args):
        raise NotImplementedError



# -*- coding: utf-8 -*-
"""
The primary model building interface in ``tinygp`` is via "kernels", which are
typically constructed as sums and products of objects defined in this
subpackage, or by subclassing :class:`Kernel` as discussed in the :ref:`kernels`
tutorial. Many of the most commonly used kernels are described in the
:ref:`stationary-kernels` section, but this section introduces some of the
fundamental building blocks.
"""

__all__ = [
    "quasisep",
    "Distance",
    "L1Distance",
    "L2Distance",
    "Kernel",
    "Conditioned",
    "Custom",
    "Sum",
    "Product",
    "Constant",
    "DotProduct",
    "Polynomial",
    "Stationary",
    "Exp",
    "ExpSquared",
    "Matern32",
    "Matern52",
    "Cosine",
    "ExpSineSquared",
    "RationalQuadratic",
]

from tinygp.kernels import quasisep
from tinygp.kernels.base import (
    Conditioned,
    Constant,
    Custom,
    DotProduct,
    Kernel,
    Polynomial,
    Product,
    Sum,
)
from tinygp.kernels.distance import Distance, L1Distance, L2Distance
from tinygp.kernels.stationary import (
    Cosine,
    Exp,
    ExpSineSquared,
    ExpSquared,
    Matern32,
    Matern52,
    RationalQuadratic,
    Stationary,
)



# -*- coding: utf-8 -*-

from __future__ import annotations

__all__ = ["DirectSolver"]

from typing import Any, Optional

import jax.numpy as jnp
import numpy as np
from jax.scipy import linalg

from tinygp import kernels
from tinygp.helpers import JAXArray, dataclass
from tinygp.noise import Noise
from tinygp.solvers.solver import Solver


@dataclass
class DirectSolver(Solver):
    """A direct solver that uses ``jax``'s built in Cholesky factorization

    You generally won't instantiate this object directly but, if you do, you'll
    probably want to use the :func:`DirectSolver.init` method instead of the
    usual constructor.
    """

    X: JAXArray
    variance_value: JAXArray
    covariance_value: JAXArray
    scale_tril: JAXArray

    @classmethod
    def init(
        cls,
        kernel: kernels.Kernel,
        X: JAXArray,
        noise: Noise,
        *,
        covariance: Optional[Any] = None,
    ) -> "DirectSolver":
        """Build a :class:`DirectSolver` for a given kernel and coordinates

        Args:
            kernel: The kernel function.
            X: The input coordinates.
            noise: The noise model for the process.
            covariance: Optionally, a pre-computed array with the covariance
                matrix. This should be equal to the result of calling ``kernel``
                and adding ``diag``, but that is not checked.
        """
        variance = kernel(X) + noise.diagonal()
        if covariance is None:
            covariance = kernel(X, X) + noise
        scale_tril = linalg.cholesky(covariance, lower=True)
        return cls(
            X=X,
            variance_value=variance,
            covariance_value=covariance,
            scale_tril=scale_tril,
        )

    def variance(self) -> JAXArray:
        return self.variance_value

    def covariance(self) -> JAXArray:
        return self.covariance_value

    def normalization(self) -> JAXArray:
        return jnp.sum(
            jnp.log(jnp.diag(self.scale_tril))
        ) + 0.5 * self.scale_tril.shape[0] * np.log(2 * np.pi)

    def solve_triangular(
        self, y: JAXArray, *, transpose: bool = False
    ) -> JAXArray:
        if transpose:
            return linalg.solve_triangular(
                self.scale_tril, y, lower=True, trans=1
            )
        else:
            return linalg.solve_triangular(self.scale_tril, y, lower=True)

    def dot_triangular(self, y: JAXArray) -> JAXArray:
        return jnp.einsum("ij,j...->i...", self.scale_tril, y)

    def condition(
        self, kernel: kernels.Kernel, X_test: Optional[JAXArray], noise: Noise
    ) -> Any:
        """Compute the covariance matrix for a conditional GP

        Args:
            kernel: The kernel for the covariance between the observed and
                predicted data.
            X_test: The coordinates of the predicted points. Defaults to the
                input coordinates.
            noise: The noise model for the predicted process.
        """
        if X_test is None:
            Ks = kernel(self.X, self.X)
            Kss = Ks + noise
        else:
            Ks = kernel(self.X, X_test)
            Kss = kernel(X_test, X_test) + noise

        A = self.solve_triangular(Ks)
        return Kss - A.transpose() @ A



# -*- coding: utf-8 -*-

from __future__ import annotations

__all__ = ["elementwise_add", "elementwise_mul", "qsm_mul"]

from typing import Optional, Tuple, TypeVar

import jax
import jax.numpy as jnp

from tinygp.helpers import JAXArray
from tinygp.solvers.quasisep.core import (
    QSM,
    DiagQSM,
    LowerTriQSM,
    SquareQSM,
    StrictLowerTriQSM,
    StrictUpperTriQSM,
    SymmQSM,
    UpperTriQSM,
)


def elementwise_add(a: QSM, b: QSM) -> Optional[QSM]:
    diag_a, lower_a, upper_a = deconstruct(a)
    diag_b, lower_b, upper_b = deconstruct(b)

    diag = add_two(diag_a, diag_b)
    lower = add_two(lower_a, lower_b)
    upper = add_two(upper_a, upper_b)

    is_symm_a = isinstance(a, SymmQSM) or isinstance(a, DiagQSM)
    is_symm_b = isinstance(b, SymmQSM) or isinstance(b, DiagQSM)
    return construct(diag, lower, upper, is_symm_a and is_symm_b)


def elementwise_mul(a: QSM, b: QSM) -> Optional[QSM]:
    diag_a, lower_a, upper_a = deconstruct(a)
    diag_b, lower_b, upper_b = deconstruct(b)

    diag = mul_two(diag_a, diag_b)
    lower = mul_two(lower_a, lower_b)
    upper = mul_two(upper_a, upper_b)

    is_symm_a = isinstance(a, SymmQSM) or isinstance(a, DiagQSM)
    is_symm_b = isinstance(b, SymmQSM) or isinstance(b, DiagQSM)
    return construct(diag, lower, upper, is_symm_a and is_symm_b)


def qsm_mul(a: QSM, b: QSM) -> Optional[QSM]:
    diag_a, lower_a, upper_a = deconstruct(a)
    diag_b, lower_b, upper_b = deconstruct(b)

    # Special case for the product of two diagonal matrices
    if (
        lower_a is None
        and upper_a is None
        and lower_b is None
        and upper_b is None
    ):
        assert diag_a is not None and diag_b is not None
        return DiagQSM(d=diag_a * diag_b)

    if lower_a is not None and upper_b is not None:

        def calc_phi(phi, data):  # type: ignore
            a, b, q, g = data
            return a @ phi @ b.T + jnp.outer(q, g), phi

        init = jnp.zeros_like(jnp.outer(lower_a.q[0], upper_b.q[0]))
        args = (lower_a.a, upper_b.a, lower_a.q, upper_b.q)
        _, phi = jax.lax.scan(calc_phi, init, args)

    else:
        phi = None

    if upper_a is not None and lower_b is not None:

        def calc_psi(psi, data):  # type: ignore
            a, b, q, g = data
            return a.T @ psi @ b + jnp.outer(q, g), psi

        init = jnp.zeros_like(jnp.outer(upper_a.q[-1], lower_b.p[-1]))
        args = (upper_a.a, lower_b.a, upper_a.p, lower_b.p)
        _, psi = jax.lax.scan(calc_psi, init, args, reverse=True)

    else:
        psi = None

    @jax.vmap
    def impl(
        diag_a: Optional[DiagQSM],
        lower_a: Optional[StrictLowerTriQSM],
        upper_a: Optional[StrictUpperTriQSM],
        diag_b: Optional[DiagQSM],
        lower_b: Optional[StrictLowerTriQSM],
        upper_b: Optional[StrictUpperTriQSM],
        phi: Optional[JAXArray],
        psi: Optional[JAXArray],
    ) -> Tuple[
        Optional[DiagQSM],
        Optional[StrictLowerTriQSM],
        Optional[StrictUpperTriQSM],
    ]:
        # Note: the order of g and h is flipped vs the paper!

        alpha = None
        beta = None
        theta = None
        eta = None
        lam = None

        if diag_b is not None and lower_a is not None:
            alpha = lower_a.q * diag_b.d

        if diag_a is not None and lower_b is not None:
            beta = diag_a.d * lower_b.p

        if diag_a is not None and upper_b is not None:
            theta = diag_a.d * upper_b.q

        if diag_b is not None and upper_a is not None:
            eta = upper_a.p * diag_b.d

        if diag_a is not None and diag_b is not None:
            lam = diag_a.d * diag_b.d

        if lower_a is not None and phi is not None and upper_b is not None:
            alpha = none_safe_add(alpha, lower_a.a @ phi @ upper_b.p)
            theta = none_safe_add(
                theta, lower_a.p @ phi @ upper_b.a.transpose()
            )
            lam = none_safe_add(lam, lower_a.p @ phi @ upper_b.p)

        if upper_a is not None and psi is not None and lower_b is not None:
            beta = none_safe_add(beta, upper_a.q @ psi @ lower_b.a)
            eta = none_safe_add(eta, upper_a.a.transpose() @ psi @ lower_b.q)
            lam = none_safe_add(lam, upper_a.q @ psi @ lower_b.q)

        s = [alpha] if alpha is not None else []
        s += [lower_b.q] if lower_b is not None else []

        t = [lower_a.p] if lower_a is not None else []
        t += [beta] if beta is not None else []

        v = [upper_a.q] if upper_a is not None else []
        v += [theta] if theta is not None else []

        u = [eta] if eta is not None else []
        u += [upper_b.p] if upper_b is not None else []

        if lower_a is not None and lower_b is not None:
            ell = jnp.concatenate(
                (
                    jnp.concatenate(
                        (lower_a.a, jnp.outer(lower_a.q, lower_b.p)), axis=-1
                    ),
                    jnp.concatenate(
                        (
                            jnp.zeros(
                                (lower_b.a.shape[0], lower_a.a.shape[0])
                            ),
                            lower_b.a,
                        ),
                        axis=-1,
                    ),
                ),
                axis=0,
            )
        else:
            ell = (
                lower_a.a
                if lower_a is not None
                else lower_b.a
                if lower_b is not None
                else None
            )

        if upper_a is not None and upper_b is not None:
            delta = jnp.concatenate(
                (
                    jnp.concatenate(
                        (
                            upper_a.a,
                            jnp.zeros(
                                (upper_a.a.shape[0], upper_b.a.shape[0])
                            ),
                        ),
                        axis=-1,
                    ),
                    jnp.concatenate(
                        (jnp.outer(upper_b.q, upper_a.p), upper_b.a), axis=-1
                    ),
                ),
                axis=0,
            )

        else:
            delta = (
                upper_a.a
                if upper_a is not None
                else upper_b.a
                if upper_b is not None
                else None
            )

        return (
            DiagQSM(d=lam) if lam is not None else None,
            StrictLowerTriQSM(
                p=jnp.concatenate(t), q=jnp.concatenate(s), a=ell
            )
            if len(t) and len(s) and ell is not None
            else None,
            StrictUpperTriQSM(
                p=jnp.concatenate(u), q=jnp.concatenate(v), a=delta
            )
            if len(u) and len(v) and delta is not None
            else None,
        )

    diag, lower, upper = impl(
        diag_a, lower_a, upper_a, diag_b, lower_b, upper_b, phi, psi
    )
    is_symm_a = isinstance(a, SymmQSM) or isinstance(a, DiagQSM)
    is_symm_b = isinstance(b, SymmQSM) or isinstance(b, DiagQSM)
    return construct(diag, lower, upper, is_symm_a and is_symm_b)


def deconstruct(
    a: QSM,
) -> Tuple[
    Optional[DiagQSM], Optional[StrictLowerTriQSM], Optional[StrictUpperTriQSM]
]:
    diag = a if isinstance(a, DiagQSM) else getattr(a, "diag", None)
    lower = (
        a if isinstance(a, StrictLowerTriQSM) else getattr(a, "lower", None)
    )
    upper = None
    if isinstance(a, StrictUpperTriQSM):
        upper = a
    elif isinstance(a, SymmQSM):
        upper = a.lower.transpose()
    elif hasattr(a, "upper"):
        upper = getattr(a, "upper")
    return diag, lower, upper


def construct(
    diag: Optional[DiagQSM],
    lower: Optional[StrictLowerTriQSM],
    upper: Optional[StrictUpperTriQSM],
    symm: bool,
) -> Optional[QSM]:
    if lower is None and upper is None:
        return diag

    if symm:
        assert diag is not None
        assert lower is not None
        return SymmQSM(diag=diag, lower=lower)

    if lower is None and upper is None:
        return diag

    if lower is None:
        if diag is None:
            return upper
        else:
            assert upper is not None
            return UpperTriQSM(diag=diag, upper=upper)

    elif upper is None:
        if diag is None:
            return lower
        else:
            assert lower is not None
            return LowerTriQSM(diag=diag, lower=lower)

    elif diag is None:
        # We would hit here if we add a StrictLower to a StrictUpper; is this an
        # ok way to handle that?
        return None

    return SquareQSM(diag=diag, lower=lower, upper=upper)


F = TypeVar("F", DiagQSM, StrictLowerTriQSM, StrictUpperTriQSM)


def add_two(a: Optional[F], b: Optional[F]) -> Optional[F]:
    if a is None and b is None:
        return None
    if a is None:
        return b
    if b is None:
        return a
    return a.self_add(b)


def mul_two(a: Optional[F], b: Optional[F]) -> Optional[F]:
    if a is None or b is None:
        return None
    return a.self_mul(b)


def none_safe_add(
    a: Optional[JAXArray], b: Optional[JAXArray]
) -> Optional[JAXArray]:
    if a is not None and b is not None:
        return a + b
    return a if a is not None else b



# -*- coding: utf-8 -*-
"""
This subpackage provides the tools needed to build expressive observation
processes for ``tinygp`` Gaussian process models. The most commonly used noise
model is :class:`Diagonal`, which adds a constant diagonal matrix to the process
covariance to represent per-observation noise. This subpackage also includes a
:class:`Dense` model for adding a full rank observation model, and
:class:`Banded` to capture noise that can be represented by a banded matrix.
"""

from __future__ import annotations

__all__ = ["Diagonal", "Dense", "Banded"]

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any, Tuple, Union

import jax.numpy as jnp
import numpy as np

from tinygp.helpers import JAXArray, dataclass

if TYPE_CHECKING:
    from tinygp.solvers.quasisep.core import DiagQSM, SymmQSM


class Noise(metaclass=ABCMeta):
    """An abstract base class defining the noise model protocol"""

    __array_priority__ = 2001

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    @abstractmethod
    def diagonal(self) -> JAXArray:
        """The diagonal elements of the noise model as an array"""
        raise NotImplementedError

    @abstractmethod
    def __add__(self, other: JAXArray) -> JAXArray:
        raise NotImplementedError

    @abstractmethod
    def __radd__(self, other: JAXArray) -> JAXArray:
        raise NotImplementedError

    @abstractmethod
    def __matmul__(self, other: JAXArray) -> JAXArray:
        raise NotImplementedError

    @abstractmethod
    def to_qsm(self) -> Union["SymmQSM", "DiagQSM"]:
        """This noise model represented as a quasiseparable matrix"""
        raise NotImplementedError


@dataclass
class Diagonal(Noise):
    """A diagonal observation noise model

    This represents the observation model using per-observation measurement
    variances.

    Args:
        diag: The diagonal elements of the noise model.
    """

    diag: JAXArray

    def __post_init__(self) -> None:
        if jnp.ndim(self.diag) != 1:
            raise ValueError(
                "The diagonal for the noise model be the same shape as the data; "
                "if passing a constant, it should be broadcasted first"
            )

    def diagonal(self) -> JAXArray:
        return self.diag

    def _add(self, other: JAXArray) -> JAXArray:
        return (
            jnp.asarray(other)
            .at[jnp.diag_indices(other.shape[0])]
            .add(self.diag)
        )

    def __add__(self, other: JAXArray) -> JAXArray:
        return self._add(other)

    def __radd__(self, other: JAXArray) -> JAXArray:
        return self._add(other)

    def __matmul__(self, other: JAXArray) -> JAXArray:
        if jnp.ndim(other) == 1:
            return self.diag * other
        else:
            return self.diag[:, None] * other

    def to_qsm(self) -> "DiagQSM":
        from tinygp.solvers.quasisep.core import DiagQSM

        return DiagQSM(d=self.diag)


@dataclass
class Dense(Noise):
    """A full rank observation noise model

    .. warning:: This model cannot be used in conjunction with the
        :class:`tinygp.solvers.QuasisepSolver` for scalable computations.

    Args:
        value: The N-by-N full rank observation model.
    """

    value: JAXArray

    def diagonal(self) -> JAXArray:
        return jnp.diag(self.value)

    def __add__(self, other: JAXArray) -> JAXArray:
        return self.value + other

    def __radd__(self, other: JAXArray) -> JAXArray:
        return other + self.value

    def __matmul__(self, other: JAXArray) -> JAXArray:
        return self.value @ other

    def to_qsm(self) -> Union["SymmQSM", "DiagQSM"]:
        """This cannot be compactly represented as a quasiseparable matrix"""
        raise NotImplementedError


@dataclass
class Banded(Noise):
    r"""A banded observation noise model

    This model captures noise that can be represented by a small number of
    off-diagonal elements in the observation matrix. One practical example of
    such an observation model is discussed by `Delisle et al. (2020)
    <https://arxiv.org/abs/2004.10678>`_. This matrix is defined by two arrays:
    ``diag`` and ``off_diags``, with shapes ``(N,)`` and ``(N, J)``
    respectively, where ``N`` is the number of data points and ``J`` is the
    number of non-zero off-diagonals required.

    For example, the following matrix has ``N = 4`` and ``J = 2``:

    .. math::

        N = \left(\begin{array}{cccc}
            n_{11} & n_{12} & n_{13} & 0      \\
            n_{12} & n_{22} & n_{23} & n_{24} \\
            n_{13} & n_{23} & n_{33} & n_{34} \\
            0      & n_{24} & n_{34} & n_{44}
        \end{array}\right)

    and it would be represented by the following arrays:

    .. code-block:: python

        diag = [n11, n22, n33, n44]

    and

    .. code-block:: python

        off_diags = [
            [n12, n13],
            [n23, n24],
            [n34,  * ],
            [ *,   * ],
        ]


    Where ``*`` represents an element that can have any arbitrary value, since it
    won't ever be accessed.
    """
    diag: JAXArray
    off_diags: JAXArray

    def diagonal(self) -> JAXArray:
        return self.diag

    def _indices(
        self,
    ) -> Tuple[Tuple[JAXArray, JAXArray], Tuple[JAXArray, JAXArray]]:
        N, J = jnp.shape(self.off_diags)
        sparse_idx_1 = []
        sparse_idx_2 = []
        dense_idx_1 = []
        dense_idx_2 = []
        for j in range(J):
            sparse_idx_1.append(np.arange(N - j - 1))
            sparse_idx_2.append(np.full(N - j - 1, j, dtype=int))
            dense_idx_1.append(np.arange(0, N - j - 1))
            dense_idx_2.append(np.arange(j + 1, N))

        return (
            (np.concatenate(sparse_idx_1), np.concatenate(sparse_idx_2)),
            (np.concatenate(dense_idx_1), np.concatenate(dense_idx_2)),
        )

    def _add(self, other: JAXArray) -> JAXArray:
        sparse_idx, dense_idx = self._indices()

        # Start by adding the diagonal
        result = (
            jnp.asarray(other)
            .at[jnp.diag_indices(other.shape[0])]
            .add(self.diag)
        )

        # Then the off diagonals, assuming symmetric
        return result.at[
            (
                np.append(dense_idx[0], dense_idx[1]),
                np.append(dense_idx[1], dense_idx[0]),
            )
        ].add(
            self.off_diags[
                (
                    np.append(sparse_idx[0], sparse_idx[0]),
                    np.append(sparse_idx[1], sparse_idx[1]),
                )
            ]
        )

    def __add__(self, other: JAXArray) -> JAXArray:
        return self._add(other)

    def __radd__(self, other: JAXArray) -> JAXArray:
        return self._add(other)

    def __matmul__(self, other: JAXArray) -> JAXArray:
        return self.to_qsm() @ other

    def to_qsm(self) -> "SymmQSM":
        from tinygp.solvers.quasisep import core

        N, J = jnp.shape(self.off_diags)
        p = jnp.repeat(jnp.eye(1, J), N, axis=0)
        q = self.off_diags
        a = jnp.repeat(jnp.eye(J, k=1)[None], N, axis=0)
        return core.SymmQSM(
            diag=core.DiagQSM(d=self.diag),
            lower=core.StrictLowerTriQSM(p=p, q=q, a=a),
        )



# -*- coding: utf-8 -*-
# mypy: ignore-errors

import jax.scipy as jsp
import numpy as np
import pytest

from tinygp.kernels import quasisep


@pytest.fixture
def random():
    return np.random.default_rng(84930)


@pytest.fixture
def data(random):
    x = np.sort(random.uniform(-3, 3, 50))
    y = np.sin(x)
    t = np.sort(random.uniform(-3, 3, 12))
    return x, y, t


@pytest.fixture(
    params=[
        quasisep.Matern32(sigma=1.8, scale=1.5),
        quasisep.Matern32(1.5),
        quasisep.Matern52(sigma=1.8, scale=1.5),
        quasisep.Matern52(1.5),
        quasisep.Celerite(1.1, 0.8, 0.9, 0.1),
        quasisep.SHO(omega=1.5, quality=0.5, sigma=1.3),
        quasisep.SHO(omega=1.5, quality=3.5, sigma=1.3),
        quasisep.SHO(omega=1.5, quality=0.1, sigma=1.3),
        quasisep.Exp(sigma=1.8, scale=1.5),
        quasisep.Exp(1.5),
        1.5 * quasisep.Matern52(1.5) + 0.3 * quasisep.Exp(1.5),
        quasisep.Matern52(1.5) * quasisep.SHO(omega=1.5, quality=0.1),
        1.5 * quasisep.Matern52(1.5) * quasisep.Celerite(1.1, 0.8, 0.9, 0.1),
        quasisep.Cosine(sigma=1.8, scale=1.5),
        1.8 * quasisep.Cosine(1.5),
        quasisep.CARMA.init(
            alpha=np.array([1.4, 2.3, 1.5]), beta=np.array([0.1, 0.5])
        ),
    ]
)
def kernel(request):
    return request.param


def test_quasisep_kernels(data, kernel):
    x, y, t = data
    K = kernel(x, x)

    # Test that to_dense and matmuls work as expected
    np.testing.assert_allclose(kernel.to_symm_qsm(x).to_dense(), K)
    np.testing.assert_allclose(kernel.matmul(x, y), K @ y)
    np.testing.assert_allclose(kernel.matmul(t, x, y), kernel(t, x) @ y)

    # Test that F and are defined consistently
    x1 = x[0]
    x2 = x[1]
    num_A = jsp.linalg.expm(kernel.design_matrix().T * (x2 - x1))
    np.testing.assert_allclose(kernel.transition_matrix(x1, x2), num_A)


def test_celerite(data):
    a, b, c, d = 1.1, 0.8, 0.9, 0.1
    kernel = quasisep.Celerite(a, b, c, d)

    x, _, t = data

    calc = kernel(x, x)
    tau = np.abs(x[:, None] - x[None, :])
    expect = np.exp(-c * tau) * (a * np.cos(d * tau) + b * np.sin(d * tau))
    np.testing.assert_allclose(calc, expect)

    calc = kernel(x, t)
    tau = np.abs(x[:, None] - t[None, :])
    expect = np.exp(-c * tau) * (a * np.cos(d * tau) + b * np.sin(d * tau))
    np.testing.assert_allclose(calc, expect)



# -*- coding: utf-8 -*-
# mypy: ignore-errors

from itertools import combinations

import jax.numpy as jnp
import numpy as np
import pytest

from tinygp.solvers.quasisep.core import (
    DiagQSM,
    LowerTriQSM,
    SquareQSM,
    StrictLowerTriQSM,
    StrictUpperTriQSM,
    SymmQSM,
)


@pytest.fixture(params=["random", "celerite"])
def name(request):
    return request.param


@pytest.fixture
def matrices(name):
    return get_matrices(name)


@pytest.fixture
def some_nice_matrices():
    diag1, p1, q1, a1, _, _, _, _ = get_matrices("celerite")
    diag2, p2, q2, a2, _, _, _, _ = get_matrices("random")
    mat1 = LowerTriQSM(
        diag=DiagQSM(diag1),
        lower=StrictLowerTriQSM(p=p1, q=q1, a=a1),
    )
    mat2 = SquareQSM(
        diag=DiagQSM(diag2),
        lower=StrictLowerTriQSM(p=p2, q=q2, a=a2),
        upper=StrictUpperTriQSM(p=p2, q=q2, a=a2),
    )
    mat3 = SquareQSM(
        diag=DiagQSM(diag1),
        lower=StrictLowerTriQSM(p=p1, q=q1, a=a1),
        upper=StrictUpperTriQSM(
            p=jnp.zeros_like(p2), q=jnp.zeros_like(q2), a=a2
        ),
    )
    mat4 = SquareQSM(
        diag=DiagQSM(diag1),
        lower=StrictLowerTriQSM(p=p1, q=q1, a=a1),
        upper=StrictUpperTriQSM(p=p2, q=q2, a=a2),
    )
    return mat1, mat2, mat3, mat4


def get_matrices(name):
    N = 100
    random = np.random.default_rng(1234)
    diag = np.exp(random.normal(size=N))

    if name == "random":
        J = 5
        p = random.normal(size=(N, J))
        q = random.normal(size=(N, J))
        a = np.repeat(np.eye(J)[None, :, :], N, axis=0)
        l = np.tril(p @ q.T, -1)
        u = np.triu(q @ p.T, 1)
        diag += np.sum(p * q, axis=1)

    elif name == "celerite":
        t = np.sort(random.uniform(0, 10, N))

        a = np.array([1.0, 2.5])
        b = np.array([0.5, 1.5])
        c = np.array([1.2, 0.5])
        d = np.array([0.5, 0.1])

        tau = np.abs(t[:, None] - t[None, :])[:, :, None]
        K = np.sum(
            np.exp(-c[None, None] * tau)
            * (
                a[None, None] * np.cos(d[None, None] * tau)
                + b[None, None] * np.sin(d[None, None] * tau)
            ),
            axis=-1,
        )
        K[np.diag_indices_from(K)] += diag
        diag = np.diag(K)
        l = np.tril(K, -1)
        u = np.triu(K, 1)

        cos = np.cos(d[None] * t[:, None])
        sin = np.sin(d[None] * t[:, None])
        p = np.concatenate(
            (
                a[None] * cos + b[None] * sin,
                a[None] * sin - b[None] * cos,
            ),
            axis=1,
        )
        q = np.concatenate((cos, sin), axis=1)
        c = np.append(c, c)
        dt = np.append(0, np.diff(t))
        a = np.stack(
            [np.diag(v) for v in np.exp(-c[None] * dt[:, None])], axis=0
        )
        p = np.einsum("ni,nij->nj", p, a)

    else:
        assert False

    v = random.normal(size=N)
    m = random.normal(size=(N, 4))
    return diag, p, q, a, v, m, l, u


def test_quasisep_def():
    random = np.random.default_rng(2022)
    n = 17
    m1 = 3
    m2 = 5
    d = random.normal(size=n)
    p = random.normal(size=(n, m1))
    q = random.normal(size=(n, m1))
    a = random.normal(size=(n, m1, m1))
    g = random.normal(size=(n, m2))
    h = random.normal(size=(n, m2))
    b = random.normal(size=(n, m2, m2))
    m = SquareQSM(
        diag=DiagQSM(d=d),
        lower=StrictLowerTriQSM(p=p, q=q, a=a),
        upper=StrictUpperTriQSM(p=g, q=h, a=b),
    ).to_dense()

    def get_value(i, j):
        if i == j:
            return d[i]
        if j < i:
            tmp = np.copy(q[j])
            for k in range(j + 1, i):
                tmp = a[k] @ tmp
            return p[i] @ tmp
        if j > i:
            tmp = np.copy(h[i])
            for k in range(i + 1, j):
                tmp = tmp @ b[k].T
            return tmp @ g[j]

    for i in range(n):
        for j in range(n):
            np.testing.assert_allclose(get_value(i, j), m[i, j])


def test_strict_tri_matmul(matrices):
    _, p, q, a, v, m, l, u = matrices
    mat = StrictLowerTriQSM(p=p, q=q, a=a)

    # Check multiplication into identity / to dense
    np.testing.assert_allclose(mat.to_dense(), l)
    np.testing.assert_allclose(mat.T.to_dense(), u)

    # Check matvec
    np.testing.assert_allclose(mat @ v, l @ v)
    np.testing.assert_allclose(mat.T @ v, u @ v)

    # Check matmat
    np.testing.assert_allclose(mat @ m, l @ m)
    np.testing.assert_allclose(mat.T @ m, u @ m)


def test_tri_matmul(matrices):
    diag, p, q, a, v, m, l, _ = matrices
    mat = LowerTriQSM(
        diag=DiagQSM(diag), lower=StrictLowerTriQSM(p=p, q=q, a=a)
    )
    dense = l + np.diag(diag)

    # Check multiplication into identity / to dense
    np.testing.assert_allclose(mat.to_dense(), dense)
    np.testing.assert_allclose(mat.T.to_dense(), dense.T)

    # Check matvec
    np.testing.assert_allclose(mat @ v, dense @ v)
    np.testing.assert_allclose(mat.T @ v, dense.T @ v)

    # Check matmat
    np.testing.assert_allclose(mat @ m, dense @ m)
    np.testing.assert_allclose(mat.T @ m, dense.T @ m)


@pytest.mark.parametrize("symm", [True, False])
def test_square_matmul(symm, matrices):
    diag, p, q, a, v, m, l, u = matrices
    if symm:
        mat = SymmQSM(
            diag=DiagQSM(diag), lower=StrictLowerTriQSM(p=p, q=q, a=a)
        )
    else:
        mat = SquareQSM(
            diag=DiagQSM(diag),
            lower=StrictLowerTriQSM(p=p, q=q, a=a),
            upper=StrictUpperTriQSM(p=p, q=q, a=a),
        )

    # Create and double check the dense reconstruction
    dense = mat.to_dense()
    np.testing.assert_allclose(np.tril(dense, -1), l)
    np.testing.assert_allclose(np.triu(dense, 1), u)
    np.testing.assert_allclose(np.diag(dense), diag)

    # Test matmuls
    np.testing.assert_allclose(mat @ v, dense @ v)
    np.testing.assert_allclose(mat @ m, dense @ m)
    np.testing.assert_allclose(v.T @ mat, v.T @ dense)
    np.testing.assert_allclose(m.T @ mat, m.T @ dense)


@pytest.mark.parametrize("name", ["celerite"])
def test_tri_inv(matrices):
    diag, p, q, a, _, _, _, _ = matrices
    mat = LowerTriQSM(
        diag=DiagQSM(diag), lower=StrictLowerTriQSM(p=p, q=q, a=a)
    )
    dense = mat.to_dense()
    minv = mat.inv()
    np.testing.assert_allclose(minv.to_dense(), jnp.linalg.inv(dense))
    np.testing.assert_allclose(
        minv.matmul(dense), np.eye(len(diag)), atol=1e-12
    )


@pytest.mark.parametrize("name", ["celerite"])
def test_tri_solve(matrices):
    diag, p, q, a, v, m, _, _ = matrices
    mat = LowerTriQSM(
        diag=DiagQSM(diag), lower=StrictLowerTriQSM(p=p, q=q, a=a)
    )
    dense = mat.to_dense()
    np.testing.assert_allclose(mat.solve(v), np.linalg.solve(dense, v))
    np.testing.assert_allclose(mat.solve(m), np.linalg.solve(dense, m))

    np.testing.assert_allclose(mat.T.solve(v), np.linalg.solve(dense.T, v))
    np.testing.assert_allclose(mat.T.solve(m), np.linalg.solve(dense.T, m))

    np.testing.assert_allclose(mat.inv().solve(v), dense @ v)
    np.testing.assert_allclose(mat.inv().solve(m), dense @ m)
    np.testing.assert_allclose(mat.T.inv().solve(v), dense.T @ v)
    np.testing.assert_allclose(mat.T.inv().solve(m), dense.T @ m)


@pytest.mark.parametrize("symm", [True, False])
def test_square_inv(symm, matrices):
    diag, p, q, a, _, _, l, u = matrices
    if symm:
        mat = SymmQSM(
            diag=DiagQSM(diag), lower=StrictLowerTriQSM(p=p, q=q, a=a)
        )
    else:
        mat = SquareQSM(
            diag=DiagQSM(diag),
            lower=StrictLowerTriQSM(p=p, q=q, a=a),
            upper=StrictUpperTriQSM(p=p, q=q, a=a),
        )

    # Create and double check the dense reconstruction
    dense = mat.to_dense()
    np.testing.assert_allclose(np.tril(dense, -1), l)
    np.testing.assert_allclose(np.triu(dense, 1), u)
    np.testing.assert_allclose(np.diag(dense), diag)

    # Invert the QS matrix
    minv = mat.inv()
    np.testing.assert_allclose(
        minv.to_dense(), jnp.linalg.inv(dense), rtol=2e-6
    )
    np.testing.assert_allclose(
        minv.matmul(dense), np.eye(len(diag)), atol=1e-12
    )

    # In this case, we know our matrix to be symmetric - so should its inverse be!
    # This may change in the future as we expand test cases
    if not symm:
        np.testing.assert_allclose(minv.lower.p, minv.upper.p)
        np.testing.assert_allclose(minv.lower.q, minv.upper.q)
        np.testing.assert_allclose(minv.lower.a, minv.upper.a)

    # The inverse of the inverse should be itself... don't actually do this!
    # Note: we can't actually directly compare the generators because there's
    # enough degrees of freedom that they won't necessarily round trip. It's
    # good enough to check that it produces the correct dense reconstruction.
    mat2 = minv.inv()
    np.testing.assert_allclose(mat2.to_dense(), dense, rtol=1e-4)


def test_gram(matrices):
    diag, p, q, a, _, _, _, _ = matrices
    mat = SquareQSM(
        diag=DiagQSM(diag),
        lower=StrictLowerTriQSM(p=p, q=q, a=a),
        upper=StrictUpperTriQSM(p=p, q=q, a=a),
    )
    dense = mat.to_dense()
    np.testing.assert_allclose(mat.gram().to_dense(), dense.T @ dense)

    mat = mat.inv()
    dense = mat.to_dense()
    np.testing.assert_allclose(mat.gram().to_dense(), dense.T @ dense)

    mat = SquareQSM(
        diag=DiagQSM(diag),
        lower=StrictLowerTriQSM(p=p, q=q, a=a),
        upper=StrictUpperTriQSM(
            p=jnp.zeros_like(p), q=jnp.zeros_like(q), a=jnp.zeros_like(a)
        ),
    )
    dense = mat.to_dense()
    np.testing.assert_allclose(mat.gram().to_dense(), dense.T @ dense)


@pytest.mark.parametrize("name", ["celerite"])
def test_cholesky(matrices):
    diag, p, q, a, v, m, _, _ = matrices
    mat = SymmQSM(diag=DiagQSM(diag), lower=StrictLowerTriQSM(p=p, q=q, a=a))
    dense = mat.to_dense()
    chol = mat.cholesky()
    np.testing.assert_allclose(chol.to_dense(), np.linalg.cholesky(dense))

    mat = mat.inv()
    dense = mat.to_dense()
    chol = mat.cholesky()
    np.testing.assert_allclose(chol.to_dense(), np.linalg.cholesky(dense))

    np.testing.assert_allclose(
        chol.solve(v), np.linalg.solve(chol.to_dense(), v)
    )
    np.testing.assert_allclose(
        chol.solve(m), np.linalg.solve(chol.to_dense(), m)
    )


def test_tri_qsmul(some_nice_matrices):
    mat1, mat2, mat3, mat4 = some_nice_matrices

    def check(mat1, mat2):
        mat = mat1 @ mat2
        a = mat.to_dense()
        b = mat1.to_dense() @ mat2.to_dense()
        np.testing.assert_allclose(np.diag(a), np.diag(b), atol=1e-12)
        np.testing.assert_allclose(np.tril(a, -1), np.tril(b, -1), atol=1e-12)
        np.testing.assert_allclose(np.triu(a, 1), np.triu(b, 1), atol=1e-12)

    minv = mat1.inv()
    mTinv = mat1.T.inv()
    for m in [mat2, mat3, mat4, mat2.inv()]:
        check(mat1, m)
        check(minv, m)
        check(mat1.T, m)
        check(mTinv, m)


def test_square_qsmul(some_nice_matrices):
    mat1, mat2, mat3, mat4 = some_nice_matrices
    mat1 += mat1.lower.transpose()

    def check(mat1, mat2):
        mat = mat1 @ mat2
        a = mat.to_dense()
        b = mat1.to_dense() @ mat2.to_dense()
        np.testing.assert_allclose(np.diag(a), np.diag(b), atol=1e-12)
        np.testing.assert_allclose(np.tril(a, -1), np.tril(b, -1), atol=1e-12)
        np.testing.assert_allclose(np.triu(a, 1), np.triu(b, 1), atol=1e-12)

    for m1, m2 in combinations(
        [mat1, mat2, mat3, mat4, mat1.inv(), mat2.inv()], 2
    ):
        check(m1, m2)


def test_ops(some_nice_matrices):
    mat1, mat2, mat3, mat4 = some_nice_matrices

    def check(mat1, mat2):
        for m1, m2 in combinations([mat1, mat2, mat1.lower, mat2.lower], 2):
            a = m1.to_dense()
            b = m2.to_dense()
            np.testing.assert_allclose((-m1).to_dense(), -a, atol=1e-12)
            np.testing.assert_allclose((m1 + m2).to_dense(), a + b, atol=1e-12)
            np.testing.assert_allclose((m1 - m2).to_dense(), a - b, atol=1e-12)
            np.testing.assert_allclose((m1 * m2).to_dense(), a * b, atol=1e-12)
            np.testing.assert_allclose(
                (2.5 * m1).to_dense(), 2.5 * a, atol=1e-12
            )

    for m1, m2 in combinations(
        [mat1, mat2, mat3, mat4, mat1.inv(), mat2.inv()], 2
    ):
        check(m1, m2)



# -*- coding: utf-8 -*-
# mypy: ignore-errors

import jax.numpy as jnp
import numpy as np

import tinygp


def check_noise_model(noise, dense_rep):
    random = np.random.default_rng(6675)

    np.testing.assert_allclose(noise.diagonal(), jnp.diag(dense_rep))
    np.testing.assert_allclose(noise + np.zeros_like(dense_rep), dense_rep)

    y1 = random.normal(size=dense_rep.shape)
    np.testing.assert_allclose(noise + y1, dense_rep + y1)
    np.testing.assert_allclose(y1 + noise, y1 + dense_rep)
    np.testing.assert_allclose(noise @ y1, dense_rep @ y1)

    y2 = random.normal(size=(dense_rep.shape[1], 3))
    np.testing.assert_allclose(noise @ y2, dense_rep @ y2)

    y3 = random.normal(size=dense_rep.shape[1])
    np.testing.assert_allclose(noise @ y3, dense_rep @ y3)

    try:
        qsm = noise.to_qsm()
    except NotImplementedError:
        pass
    else:
        np.testing.assert_allclose(qsm @ y1, dense_rep @ y1)
        np.testing.assert_allclose(qsm @ y2, dense_rep @ y2)
        np.testing.assert_allclose(qsm @ y3, dense_rep @ y3)


def test_diagonal():
    N = 50
    random = np.random.default_rng(9432)
    diag = random.normal(size=N)
    noise = tinygp.noise.Diagonal(diag=diag)
    check_noise_model(noise, np.diag(diag))


def test_banded():
    N, J = 50, 5
    random = np.random.default_rng(9432)

    # Create a random symmetric banded matrix
    R = random.normal(size=(N, N))
    R[np.triu_indices(N, J + 1)] = 0
    R[np.tril_indices(N)] = R.T[np.tril_indices(N)]

    # Extract the diagonal and off-diagonal elements
    diag = np.diag(R)
    off_diags = np.zeros((N, J))
    for j in range(J):
        off_diags[: N - j - 1, j] = R[
            (np.arange(0, N - j - 1), np.arange(j + 1, N))
        ]

    noise = tinygp.noise.Banded(diag=diag, off_diags=off_diags)
    check_noise_model(noise, R)


def test_dense():
    N = 50
    random = np.random.default_rng(9432)
    M = random.normal(size=(N, N))
    noise = tinygp.noise.Dense(value=M)
    check_noise_model(noise, M)



# -*- coding: utf-8 -*-
"""
In ``tinygp``, the Gaussian process mean function can be defined using any
callable object, but this submodule includes two helper classes for defining
means. When defining your own mean function, it's important to remember that
your callable should accept as input a single input coordinate (i.e. not a
*vector* of coordinates), and return the scalar value of the mean at that
coordinate. ``tinygp`` will handle all the relevant ``vmap``-ing and
broadcasting.
"""

from __future__ import annotations

__all__ = ["Mean", "Conditioned"]

from typing import TYPE_CHECKING, Any, Callable, Optional, Union

import jax

from tinygp.helpers import JAXArray, dataclass
from tinygp.kernels.base import Kernel


@dataclass
class Mean:
    """A wrapper for the GP mean which supports a constant value or a callable

    In ``tinygp``, a mean function can be any callable which takes as input a
    single coordinate and returns the scalar mean at that location.

    Args:
        value: Either a *scalar* constant, or a callable with the correct
            signature.
    """

    value: Union[JAXArray, Callable[[JAXArray], JAXArray]]

    if TYPE_CHECKING:

        def __init__(self, *args: Any, **kwargs: Any) -> None:
            pass

    def __call__(self, X: JAXArray) -> JAXArray:
        if callable(self.value):
            return self.value(X)
        return self.value


@dataclass
class Conditioned:
    """The mean of a process conditioned on observed data

    Args:
        X: The coordinates of the data. alpha: The value :math:`L^-1\,y` where L
        is ``scale_tril`` and y is the
            observed data.
        scale_tril: The lower Cholesky factor of the base process' kernel
            matrix.
        kernel: The predictive kerenl; this will generally be the kernel from
            the kernel used by the original process.
        include_mean: If ``True``, the predicted values will include the mean
            function evaluated at ``X_test``.
        mean_function: The mean function of the base process. Used only if
            ``include_mean`` is ``True``.
    """

    X: JAXArray
    alpha: JAXArray
    kernel: Kernel
    include_mean: bool
    mean_function: Optional[Mean] = None

    if TYPE_CHECKING:

        def __init__(self, *args: Any, **kwargs: Any) -> None:
            pass

    def __call__(self, X: JAXArray) -> JAXArray:
        Ks = jax.vmap(self.kernel.evaluate, in_axes=(None, 0), out_axes=0)(
            X, self.X
        )
        mu = Ks @ self.alpha
        if self.include_mean and self.mean_function is not None:
            mu += self.mean_function(X)
        return mu



# -*- coding: utf-8 -*-
"""
While the usual definition of quasiseparable matrices is restricted to square
matrices, it is useful for our purposes to also implement some algorithms for a
somewhat more general class of rectangular quasiseparable matrices. These appear
in the calculations for the conditional Gaussian Process when interpolating and
extrapolating. We have not (yet?) worked through some of the more general
operations (like scalable matrix multiplies), but those may be possible to
derive.

This generalization isn't published anywhere as far as I know (please tell me if
there is a reference that you know of!), and maybe someday I'll come up with
notation that I'm satisfied with and try to write it up.
"""

from __future__ import annotations

__all__ = ["GeneralQSM"]

from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Tuple

import jax
import jax.numpy as jnp

from tinygp.helpers import JAXArray, dataclass


def handle_matvec_shapes(
    func: Callable[[Any, JAXArray], JAXArray]
) -> Callable[[Any, JAXArray], JAXArray]:
    @wraps(func)
    def wrapped(self: Any, x: JAXArray) -> JAXArray:
        output_shape = (-1,) + x.shape[1:]
        result = func(self, jnp.reshape(x, (x.shape[0], -1)))
        return jnp.reshape(result, output_shape)

    return wrapped


@dataclass
class GeneralQSM:
    """A rectangular ``(n1,n2)`` quasiseparable matrix with order ``m``

    Args:
        pl (n1, m): The lower left quasiseparable vectors.
        ql (n2, m): The lower right quasiseparable vectors.
        pu (n2, m): The upper right quasiseparable vectors.
        qu (n1, m): The upper left quasiseparable vectors.
        a (n1, m, m): The transition matrices.
        idx (n1,): The indices of the diagonal.
    """

    pl: JAXArray
    ql: JAXArray
    pu: JAXArray
    qu: JAXArray
    a: JAXArray
    idx: JAXArray

    if TYPE_CHECKING:

        def __init__(self, *args: Any, **kwargs: Any) -> None:
            pass

    @property
    def shape(self) -> Tuple[int, int]:
        """The shape of the matrix"""
        return (self.pl.shape[0], self.ql.shape[0])

    @jax.jit
    @handle_matvec_shapes
    def matmul(self, x: JAXArray) -> JAXArray:
        """The dot product of this matrix with a dense vector or matrix

        Args:
            x (n2, ...): A matrix or vector with leading dimension matching this
                matrix.
        """

        # Use a forward pass to dot the "lower" matrix
        def forward(f, data):  # type: ignore
            q, a, x = data
            fn = a @ f + jnp.outer(q, x)
            return fn, fn

        init = jnp.zeros_like(jnp.outer(self.ql[0], x[0]))
        _, f = jax.lax.scan(forward, init, (self.ql, self.a, x))
        idx = jnp.clip(self.idx, 0, f.shape[0] - 1)
        mask = jnp.logical_and(self.idx >= 0, self.idx < f.shape[0])
        lower = jax.vmap(jnp.dot)(jnp.where(mask[:, None], self.pl, 0), f[idx])

        # Then a backward pass to apply the "upper" matrix
        def backward(f, data):  # type: ignore
            p, a, x = data
            fn = a.T @ f + jnp.outer(p, x)
            return fn, fn

        init = jnp.zeros_like(jnp.outer(self.pu[-1], x[-1]))
        _, f = jax.lax.scan(
            backward,
            init,
            (self.pu, jnp.roll(self.a, -1, axis=0), x),
            reverse=True,
        )
        idx = jnp.clip(self.idx + 1, 0, f.shape[0] - 1)
        mask = jnp.logical_and(self.idx >= -1, self.idx < f.shape[0] - 1)
        upper = jax.vmap(jnp.dot)(jnp.where(mask[:, None], self.qu, 0), f[idx])

        return lower + upper

    def __matmul__(self, other: Any) -> Any:
        return self.matmul(other)



# -*- coding: utf-8 -*-
# mypy: ignore-errors

import jax.numpy as jnp
import numpy as np

from tinygp import kernels, transforms


def test_linear():
    kernel0 = kernels.Matern32(4.5)
    kernel1 = transforms.Linear(1 / 4.5, kernels.Matern32())
    np.testing.assert_allclose(
        kernel0.evaluate(0.5, 0.1), kernel1.evaluate(0.5, 0.1)
    )


def test_multivariate_linear():
    kernel0 = kernels.Matern32(4.5)
    kernel1 = transforms.Linear(jnp.full(3, 1 / 4.5), kernels.Matern32())
    np.testing.assert_allclose(
        kernel0.evaluate(jnp.full(3, 0.5), jnp.full(3, 0.1)),
        kernel1.evaluate(jnp.full(3, 0.5), jnp.full(3, 0.1)),
    )


def test_cholesky():
    kernel0 = kernels.Matern32(4.5)
    kernel1 = transforms.Cholesky(4.5, kernels.Matern32())
    np.testing.assert_allclose(
        kernel0.evaluate(0.5, 0.1), kernel1.evaluate(0.5, 0.1)
    )


def test_multivariate_cholesky():
    kernel0 = kernels.Matern32(4.5)
    kernel1 = transforms.Cholesky(jnp.full(3, 4.5), kernels.Matern32())
    np.testing.assert_allclose(
        kernel0.evaluate(jnp.full(3, 0.5), jnp.full(3, 0.1)),
        kernel1.evaluate(jnp.full(3, 0.5), jnp.full(3, 0.1)),
    )


def test_subspace():
    kernel = transforms.Subspace(1, kernels.Matern32())
    np.testing.assert_allclose(
        kernel.evaluate(np.array([0.5, 0.1]), np.array([-0.4, 0.7])),
        kernel.evaluate(np.array([100.5, 0.1]), np.array([-70.4, 0.7])),
    )



# -*- coding: utf-8 -*-
r"""
This subpackage implements the core linear algebra required for the
:class:`QuasisepSolver`, and a few extras. An intro to these matrices can be
found in `Matrix Computations and Semiseparable Matrices: Linear Systems
<https://muse.jhu.edu/book/16537>`_.

There exist a range of definitions for *quasiseparable matrices* in the
literature, so to be explicit, let's select the one that we will consider in all
that follows. The most suitable definition for our purposes is nearly identical
to the one used by `Eidelman & Gohberg (1999)
<https://link.springer.com/article/10.1007%2FBF01300581>`_, with some small
modifications.

Let's start by considering an :math:`N \times N` *square quasiseparable matrix*
:math:`M` with lower quasiseparable order :math:`m_l` and upper quasiseparable
order :math:`m_u`. We represent this matrix :math:`M` as:

.. math::

  M_{ij} = \left \{ \begin{array}{ll}
    d_i\quad, & \mbox{if }\, i = j \\ {p}_i^T\,\left ( \prod_{k=i-1}^{j+1} {A}_k
    \right )\,{q}_j\quad,   & \mbox{if }\, i > j \\ {h}_i^T\,\left (
    \prod_{k=i+1}^{j-1} {B}_k^T \right )\,{g}_j\quad, & \mbox{if }\, i < j \\
  \end{array}\right .

where

- :math:`i` and :math:`j` both range from :math:`1` to :math:`N`,
- :math:`d_i` is a scalar,
- :math:`{p}_i` and :math:`{q}_j` are both vectors with :math:`m_l` elements,
- :math:`{A}_k` is an :math:`m_l \times m_l` matrix,
- :math:`{g}_j` and :math:`{h}_i` are both vectors with :math:`m_u` elements,
  and
- :math:`{B}_k` is an :math:`m_u \times m_u` matrix.

Comparing this definition to the one from Eidelman & Gohberg, you may notice
that we have swapped the labels of :math:`{g}_j` and :math:`{h}_i`, and that
we've added an explicit transpose to :math:`{B}_k^T`. These changes simplify the
notation and implementation for symmetric matrices where, with our definition,
:math:`{g} = {p}`, :math:`{h} = {q}`, and :math:`{B} = {A}`.

In our definition, the product notation is a little sloppy so, to be more
explicit, this is how the products expand:

.. math::

  \prod_{k=i-1}^{j+1} {A}_k \equiv
  {A}_{i-1}\,{A}_{i-2}\cdots{A}_{j+2}\,{A}_{j+1}

and

.. math::

  \prod_{k=i+1}^{j-1} {B}_k^T \equiv
  {B}_{i+1}^T\,{B}_{i+2}^T\cdots{B}_{j-2}^T\,{B}_{j-1}^T

It is often useful see a concrete example matrix. For example, under our
definition, the general :math:`4 \times 4` quasiseparable matrix is:

.. math::

    M = \left(\begin{array}{cccc}
      d_1 & {h}_1^T{g}_2 & {h}_1^T{B}_2^T{g}_3 & {h}_1^T{B}_2^T{B}_3^T{g}_4 \\
      {p}_2^T{q}_1 & d_2 & {h}_2^T{g}_3 & {h}_2^T{B}_3^T{g}_4 \\
      {p}_3^T{A}_2{q}_1 & {p}_3^T {q}_2 & d_3 & {h}_3^T{g}_4 \\
      {p}_4^T{A}_3{A}_2{q}_1 & {p}_4^T{A}_3{q}_2 & {p}_4^T{q}_3 & d_4 \\
    \end{array}\right)

These matrices allow linear scaling for most basic linear algebra operations
(that's why we like them!), and this subpackage implements many of these
building blocks.
"""

__all__ = ["QuasisepSolver"]

from tinygp.solvers.quasisep.solver import QuasisepSolver



# -*- coding: utf-8 -*-

from __future__ import annotations

__all__ = ["JAXArray", "dataclass", "field"]

import dataclasses
from typing import Any, Callable, Tuple, Type, TypeVar, Union

import jax
import jax.numpy as jnp
import numpy as np

JAXArray = Union[np.ndarray, jnp.ndarray]

# This section is based closely on the implementation in flax:
#
# https://github.com/google/flax/blob/b60f7f45b90f8fc42a88b1639c9cc88a40b298d3/flax/struct.py
#
# This decorator is interpreted by static analysis tools as a hint
# that a decorator or metaclass causes dataclass-like behavior.
# See https://github.com/microsoft/pyright/blob/main/specs/dataclass_transforms.md
# for more information about the __dataclass_transform__ magic.
_T = TypeVar("_T")


def __dataclass_transform__(
    *,
    eq_default: bool = True,
    order_default: bool = False,
    kw_only_default: bool = False,
    field_descriptors: Tuple[Union[type, Callable[..., Any]], ...] = (()),
) -> Callable[[_T], _T]:
    # If used within a stub file, the following implementation can be
    # replaced with "...".
    return lambda a: a


@__dataclass_transform__()
def dataclass(clz: Type[Any]) -> Type[Any]:
    data_clz: Any = dataclasses.dataclass(frozen=True)(clz)
    meta_fields = []
    data_fields = []
    for name, field_info in data_clz.__dataclass_fields__.items():
        is_pytree_node = field_info.metadata.get("pytree_node", True)
        if is_pytree_node:
            data_fields.append(name)
        else:
            meta_fields.append(name)

    def replace(self: Any, **updates: _T) -> _T:
        return dataclasses.replace(self, **updates)

    data_clz.replace = replace

    def iterate_clz(x: Any) -> Tuple[Tuple[Any, ...], Tuple[Any, ...]]:
        meta = tuple(getattr(x, name) for name in meta_fields)
        data = tuple(getattr(x, name) for name in data_fields)
        return data, meta

    def clz_from_iterable(meta: Tuple[Any, ...], data: Tuple[Any, ...]) -> Any:
        meta_args = tuple(zip(meta_fields, meta))
        data_args = tuple(zip(data_fields, data))
        kwargs = dict(meta_args + data_args)
        return data_clz(**kwargs)

    jax.tree_util.register_pytree_node(
        data_clz, iterate_clz, clz_from_iterable
    )

    # Hack to make this class act as a tuple when unpacked
    data_clz.iter_elems = lambda self: iterate_clz(self)[0].__iter__()

    return data_clz


def field(pytree_node: bool = True, **kwargs: Any) -> Any:
    return dataclasses.field(metadata={"pytree_node": pytree_node}, **kwargs)



# -*- coding: utf-8 -*-
# mypy: ignore-errors

import warnings

import numpy as np
import pytest

from tinygp import GaussianProcess, kernels

george = pytest.importorskip("george")


@pytest.fixture
def random():
    return np.random.default_rng(1058390)


@pytest.fixture(
    scope="module",
    params=["Constant", "DotProduct", "Polynomial"],
)
def kernel(request):
    return {
        "Constant": (
            kernels.Constant(value=1.5),
            george.kernels.ConstantKernel(
                log_constant=np.log(1.5 / 5), ndim=5
            ),
        ),
        "DotProduct": (
            kernels.DotProduct(),
            george.kernels.DotProductKernel(ndim=3),
        ),
        "Polynomial": (
            kernels.Polynomial(order=2.5, sigma=1.3),
            george.kernels.PolynomialKernel(
                order=2.5, log_sigma2=2 * np.log(1.3), ndim=1
            ),
        ),
    }[request.param]


@pytest.fixture(scope="module", params=["Cosine", "ExpSineSquared"])
def periodic_kernel(request):
    return {
        "Cosine": (
            kernels.Cosine(scale=2.3),
            george.kernels.CosineKernel(log_period=np.log(2.3)),
        ),
        "ExpSineSquared": (
            kernels.ExpSineSquared(scale=2.3, gamma=1.3),
            george.kernels.ExpSine2Kernel(gamma=1.3, log_period=np.log(2.3)),
        ),
    }[request.param]


@pytest.fixture(
    scope="module",
    params=["Exp", "ExpSquared", "Matern32", "Matern52", "RationalQuadratic"],
)
def stationary_kernel(request):
    scale = 1.5
    return {
        "Exp": (
            kernels.Exp(scale),
            george.kernels.ExpKernel(scale**2),
        ),
        "ExpSquared": (
            kernels.ExpSquared(scale),
            george.kernels.ExpSquaredKernel(scale**2),
        ),
        "Matern32": (
            kernels.Matern32(scale),
            george.kernels.Matern32Kernel(scale**2),
        ),
        "Matern52": (
            kernels.Matern52(scale),
            george.kernels.Matern52Kernel(scale**2),
        ),
        "RationalQuadratic": (
            kernels.RationalQuadratic(alpha=1.5),
            george.kernels.RationalQuadraticKernel(
                metric=1.0, log_alpha=np.log(1.5)
            ),
        ),
    }[request.param]


def compare_kernel_value(random, tiny_kernel, george_kernel):
    x1 = np.sort(random.uniform(0, 10, (50, george_kernel.ndim)))
    x2 = np.sort(random.uniform(0, 10, (45, george_kernel.ndim)))
    np.testing.assert_allclose(
        tiny_kernel(x1, x2),
        george_kernel.get_value(x1, x2),
    )
    np.testing.assert_allclose(
        tiny_kernel(x1, x1),
        george_kernel.get_value(x1),
    )
    np.testing.assert_allclose(
        tiny_kernel(x1),
        george_kernel.get_value(x1, diag=True),
    )


def compare_gps(random, tiny_kernel, george_kernel):
    x = np.sort(random.uniform(0, 10, (50, george_kernel.ndim)))
    t = np.sort(random.uniform(0, 10, (12, george_kernel.ndim)))
    y = np.sin(x[:, 0])
    diag = random.uniform(0.01, 0.1, 50)

    # Set up the GPs
    george_gp = george.GP(george_kernel)
    george_gp.compute(x, np.sqrt(diag))
    tiny_gp = GaussianProcess(tiny_kernel, x, diag=diag)

    # Likelihood
    np.testing.assert_allclose(
        tiny_gp.log_probability(y), george_gp.log_likelihood(y)
    )

    # Filtering
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)

        np.testing.assert_allclose(
            tiny_gp.predict(y),
            george_gp.predict(y, x, return_var=False, return_cov=False),
        )

        # Filtering with explicit value
        np.testing.assert_allclose(
            tiny_gp.predict(y, x),
            george_gp.predict(y, x, return_var=False, return_cov=False),
        )
        np.testing.assert_allclose(
            tiny_gp.predict(y, t),
            george_gp.predict(y, t, return_var=False, return_cov=False),
        )

        # Variance
        np.testing.assert_allclose(
            tiny_gp.predict(y, return_var=True)[1],
            george_gp.predict(y, x, return_var=True, return_cov=False)[1],
            atol=1e-7,
        )
        np.testing.assert_allclose(
            tiny_gp.predict(y, t, return_var=True)[1],
            george_gp.predict(y, t, return_var=True, return_cov=False)[1],
            atol=1e-7,
        )

        # Covariance
        np.testing.assert_allclose(
            tiny_gp.predict(y, return_cov=True)[1],
            george_gp.predict(y, x, return_var=False, return_cov=True)[1],
            atol=1e-7,
        )
        np.testing.assert_allclose(
            tiny_gp.predict(y, t, return_cov=True)[1],
            george_gp.predict(y, t, return_var=False, return_cov=True)[1],
            atol=1e-7,
        )


def test_kernel_value(random, kernel):
    tiny_kernel, george_kernel = kernel
    compare_kernel_value(random, tiny_kernel, george_kernel)
    tiny_kernel *= 0.3
    george_kernel *= 0.3
    compare_kernel_value(random, tiny_kernel, george_kernel)


def test_periodic_kernel_value(random, periodic_kernel):
    tiny_kernel, george_kernel = periodic_kernel
    compare_kernel_value(random, tiny_kernel, george_kernel)
    tiny_kernel *= 0.3
    george_kernel *= 0.3
    compare_kernel_value(random, tiny_kernel, george_kernel)


def test_metric_kernel_value(random, stationary_kernel):
    tiny_kernel, george_kernel = stationary_kernel
    compare_kernel_value(random, tiny_kernel, george_kernel)
    tiny_kernel *= 0.3
    george_kernel *= 0.3
    compare_kernel_value(random, tiny_kernel, george_kernel)


def test_gp(random, kernel):
    tiny_kernel, george_kernel = kernel
    compare_gps(random, tiny_kernel, george_kernel)


def test_periodic_gp(random, periodic_kernel):
    tiny_kernel, george_kernel = periodic_kernel
    compare_gps(random, tiny_kernel, george_kernel)


def test_metric_gp(random, stationary_kernel):
    tiny_kernel, george_kernel = stationary_kernel
    compare_gps(random, tiny_kernel, george_kernel)



# -*- coding: utf-8 -*-
"""
This submodule defines a set of distance metrics that can be used when working
with multivariate data. By default, all
:class:`tinygp.kernels.stationary.Stationary` kernels will use either an
:class:`L1Distance` or :class:`L2Distance`, when applied in multiple dimensions,
but it is possible to define custom metrics, as discussed in the :ref:`geometry`
tutorial.
"""

from __future__ import annotations

__all__ = ["Distance", "L1Distance", "L2Distance"]

from abc import ABCMeta, abstractmethod

import jax.numpy as jnp

from tinygp.helpers import JAXArray, dataclass


class Distance(metaclass=ABCMeta):
    """An abstract base class defining a distance metric interface"""

    @abstractmethod
    def distance(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """Compute the distance between two coordinates under this metric"""
        raise NotImplementedError()

    def squared_distance(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """Compute the squared distance between two coordinates

        By default this returns the squared result of
        :func:`tinygp.kernels.stationary.Distance.distance`, but some metrics
        can take advantage of these separate implementations to avoid
        unnecessary square roots.
        """
        return jnp.square(self.distance(X1, X2))


@dataclass
class L1Distance(Distance):
    """The L1 or Manhattan distance between two coordinates"""

    def distance(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        return jnp.sum(jnp.abs(X1 - X2))


@dataclass
class L2Distance(Distance):
    """The L2 or Euclidean distance between two coordinates"""

    def distance(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        r1 = L1Distance().distance(X1, X2)
        r2 = self.squared_distance(X1, X2)
        zeros = jnp.equal(r2, 0)
        r2 = jnp.where(zeros, jnp.ones_like(r2), r2)
        return jnp.where(zeros, r1, jnp.sqrt(r2))

    def squared_distance(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        return jnp.sum(jnp.square(X1 - X2))



# -*- coding: utf-8 -*-
# mypy: ignore-errors

import jax
import jax.numpy as jnp
import numpy as np
from jax.test_util import check_grads

from tinygp.kernels import distance


def check(comp, expect, args, order=2, **kwargs):
    np.testing.assert_allclose(expect(*args), comp(*args))
    np.testing.assert_allclose(jax.grad(comp)(*args), jax.grad(expect)(*args))
    check_grads(comp, args, order=order, **kwargs)


def test_l2_distance_grad_at_zero():
    expect = lambda x1, x2: jnp.sqrt(jnp.sum(jnp.square(x1 - x2)))
    comp = distance.L2Distance().distance

    x1 = 0.0
    x2 = 1.5
    check(comp, expect, (x1, x2))

    x1 = jnp.array([0.0, 0.1])
    x2 = jnp.array([1.5, -0.2])
    check(comp, expect, (x1, x2))

    g = jax.grad(comp)(x1, x1)
    np.testing.assert_allclose(expect(x1, x1), comp(x1, x1))
    assert np.all(np.isfinite(g))



# -*- coding: utf-8 -*-

from __future__ import annotations

__all__ = ["GaussianProcess"]

from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import jax
import jax.numpy as jnp

from tinygp import kernels, means
from tinygp.helpers import JAXArray
from tinygp.kernels.quasisep import Quasisep
from tinygp.noise import Diagonal, Noise
from tinygp.solvers import DirectSolver, QuasisepSolver
from tinygp.solvers.quasisep.core import SymmQSM

if TYPE_CHECKING:
    from tinygp.numpyro_support import TinyDistribution


class GaussianProcess:
    """An interface for designing a Gaussian Process regression model

    Args:
        kernel (Kernel): The kernel function
        X (JAXArray): The input coordinates. This can be any PyTree that is
            compatible with ``kernel`` where the zeroth dimension is ``N_data``,
            the size of the data set.
        diag (JAXArray, optional): The value to add to the diagonal of the
            covariance matrix, often used to capture measurement uncertainty.
            This should be a scalar or have the shape ``(N_data,)``. If not
            provided, this will default to the square root of machine epsilon
            for the data type being used. This can sometimes be sufficient to
            avoid numerical issues, but if you're getting NaNs, try increasing
            this value.
        noise (Noise, optional): Used to implement more expressive observation
            noise models than those supported by just ``diag``. This can be any
            object that implements the :class:`tinygp.noise.Noise` protocol. If
            this is provided, the ``diag`` parameter will be ignored.
        mean (Callable, optional): A callable or constant mean function that
            will be evaluated with the ``X`` as input: ``mean(X)``
        solver: The solver type to be used to execute the required linear
            algebra.
    """

    def __init__(
        self,
        kernel: kernels.Kernel,
        X: JAXArray,
        *,
        diag: Optional[JAXArray] = None,
        noise: Optional[Noise] = None,
        mean: Optional[Union[Callable[[JAXArray], JAXArray], JAXArray]] = None,
        solver: Optional[Any] = None,
        mean_value: Optional[JAXArray] = None,
        covariance_value: Optional[Any] = None,
        **solver_kwargs: Any,
    ):
        self.kernel = kernel
        self.X = X

        if callable(mean):
            self.mean_function = mean
        else:
            if mean is None:
                self.mean_function = means.Mean(jnp.zeros(()))
            else:
                self.mean_function = means.Mean(mean)
        if mean_value is None:
            mean_value = jax.vmap(self.mean_function)(self.X)
        self.num_data = mean_value.shape[0]
        self.dtype = mean_value.dtype
        self.loc = self.mean = mean_value
        if self.mean.ndim != 1:
            raise ValueError(
                "Invalid mean shape: "
                f"expected ndim = 1, got ndim={self.mean.ndim}"
            )

        if noise is None:
            diag = _default_diag(self.mean) if diag is None else diag
            noise = Diagonal(diag=jnp.broadcast_to(diag, self.mean.shape))
        self.noise = noise

        if solver is None:
            if isinstance(covariance_value, SymmQSM) or isinstance(
                kernel, Quasisep
            ):
                solver = QuasisepSolver
            else:
                solver = DirectSolver
        self.solver = solver.init(
            kernel,
            self.X,
            self.noise,
            covariance=covariance_value,
            **solver_kwargs,
        )

    @property
    def variance(self) -> JAXArray:
        return self.solver.variance()

    @property
    def covariance(self) -> JAXArray:
        return self.solver.covariance()

    def log_probability(self, y: JAXArray) -> JAXArray:
        """Compute the log probability of this multivariate normal

        Args:
            y (JAXArray): The observed data. This should have the shape
                ``(N_data,)``, where ``N_data`` was the zeroth axis of the ``X``
                data provided when instantiating this object.

        Returns:
            The marginal log probability of this multivariate normal model,
            evaluated at ``y``.
        """
        return self._compute_log_prob(self._get_alpha(y))

    def condition(
        self,
        y: JAXArray,
        X_test: Optional[JAXArray] = None,
        *,
        diag: Optional[JAXArray] = None,
        noise: Optional[Noise] = None,
        include_mean: bool = True,
        kernel: Optional[kernels.Kernel] = None,
    ) -> ConditionResult:
        """Condition the model on observed data and

        Args:
            y (JAXArray): The observed data. This should have the shape
                ``(N_data,)``, where ``N_data`` was the zeroth axis of the ``X``
                data provided when instantiating this object.
            X_test (JAXArray, optional): The coordinates where the prediction
                should be evaluated. This should have a data type compatible
                with the ``X`` data provided when instantiating this object. If
                it is not provided, ``X`` will be used by default, so the
                predictions will be made.
            diag (JAXArray, optional): Will be passed as the diagonal to the
                conditioned ``GaussianProcess`` object, so this can be used to
                introduce, for example, observational noise to predicted data.
            include_mean (bool, optional): If ``True`` (default), the predicted
                values will include the mean function evaluated at ``X_test``.
            kernel (Kernel, optional): A kernel to optionally specify the
                covariance between the observed data and predicted data. See
                :ref:`mixture` for an example.

        Returns:
            A named tuple where the first element ``log_probability`` is the log
            marginal probability of the model, and the second element ``gp`` is
            the :class:`GaussianProcess` object describing the conditional
            distribution evaluated at ``X_test``.
        """
        # If X_test is provided, we need to check that the tree structure
        # matches that of the input data, and that the shapes are all compatible
        # (i.e. the dimension of the inputs must match). This is slightly
        # convoluted since we need to support arbitrary pytrees.
        if X_test is not None:
            matches = jax.tree_util.tree_map(
                lambda a, b: jnp.ndim(a) == jnp.ndim(b)
                and jnp.shape(a)[1:] == jnp.shape(b)[1:],
                self.X,
                X_test,
            )
            if not jax.tree_util.tree_reduce(lambda a, b: a and b, matches):
                raise ValueError(
                    "`X_test` must have the same tree structure as the input `X`, "
                    "and all but the leading dimension must have matching sizes"
                )

        alpha, log_prob, mean_value = self._condition(
            y, X_test, include_mean, kernel
        )
        if kernel is None:
            kernel = self.kernel

        if noise is None:
            diag = _default_diag(mean_value) if diag is None else diag
            noise = Diagonal(diag=jnp.broadcast_to(diag, mean_value.shape))

        covariance_value = self.solver.condition(kernel, X_test, noise)
        if X_test is None:
            X_test = self.X

        # The conditional GP will also be a GP with the mean an covariance
        # specified by a :class:`tinygp.means.Conditioned` and
        # :class:`tinygp.kernels.Conditioned` respectively.
        gp = GaussianProcess(
            kernels.Conditioned(self.X, self.solver, kernel),
            X_test,
            noise=noise,
            mean=means.Conditioned(
                self.X,
                alpha,
                kernel,
                include_mean=include_mean,
                mean_function=self.mean_function,
            ),
            mean_value=mean_value,
            covariance_value=covariance_value,
        )

        return ConditionResult(log_prob, gp)

    @partial(
        jax.jit,
        static_argnums=(0,),
        static_argnames=("include_mean", "return_var", "return_cov"),
    )
    def predict(
        self,
        y: JAXArray,
        X_test: Optional[JAXArray] = None,
        *,
        kernel: Optional[kernels.Kernel] = None,
        include_mean: bool = True,
        return_var: bool = False,
        return_cov: bool = False,
    ) -> Union[JAXArray, Tuple[JAXArray, JAXArray]]:
        """Predict the GP model at new test points conditioned on observed data

        Args:
            y (JAXArray): The observed data. This should have the shape
                ``(N_data,)``, where ``N_data`` was the zeroth axis of the ``X``
                data provided when instantiating this object.
            X_test (JAXArray, optional): The coordinates where the prediction
                should be evaluated. This should have a data type compatible
                with the ``X`` data provided when instantiating this object. If
                it is not provided, ``X`` will be used by default, so the
                predictions will be made.
            include_mean (bool, optional): If ``True`` (default), the predicted
                values will include the mean function evaluated at ``X_test``.
            return_var (bool, optional): If ``True``, the variance of the
                predicted values at ``X_test`` will be returned.
            return_cov (bool, optional): If ``True``, the covariance of the
                predicted values at ``X_test`` will be returned. If
                ``return_var`` is ``True``, this flag will be ignored.

        Returns:
            The mean of the predictive model evaluated at ``X_test``, with shape
            ``(N_test,)`` where ``N_test`` is the zeroth dimension of
            ``X_test``. If either ``return_var`` or ``return_cov`` is ``True``,
            the variance or covariance of the predicted process will also be
            returned with shape ``(N_test,)`` or ``(N_test, N_test)``
            respectively.
        """
        _, cond = self.condition(
            y, X_test, kernel=kernel, include_mean=include_mean
        )
        if return_var:
            return cond.loc, cond.variance
        if return_cov:
            return cond.loc, cond.covariance
        return cond.loc

    def sample(
        self,
        key: jax.random.KeyArray,
        shape: Optional[Sequence[int]] = None,
    ) -> JAXArray:
        """Generate samples from the prior process

        Args:
            key: A ``jax`` random number key array. shape (tuple, optional): The
            number and shape of samples to
                generate.

        Returns:
            The sampled realizations from the process with shape ``(N_data,) +
            shape`` where ``N_data`` is the zeroth dimension of the ``X``
            coordinates provided when instantiating this process.
        """
        return self._sample(key, shape)

    def numpyro_dist(self, **kwargs: Any) -> "TinyDistribution":
        """Get the numpyro MultivariateNormal distribution for this process"""
        from tinygp.numpyro_support import TinyDistribution

        return TinyDistribution(self, **kwargs)

    @partial(jax.jit, static_argnums=(0, 2))
    def _sample(
        self,
        key: jax.random.KeyArray,
        shape: Optional[Sequence[int]],
    ) -> JAXArray:
        if shape is None:
            shape = (self.num_data,)
        else:
            shape = (self.num_data,) + tuple(shape)
        normal_samples = jax.random.normal(key, shape=shape, dtype=self.dtype)
        return self.mean + jnp.moveaxis(
            self.solver.dot_triangular(normal_samples), 0, -1
        )

    @partial(jax.jit, static_argnums=0)
    def _compute_log_prob(self, alpha: JAXArray) -> JAXArray:
        loglike = (
            -0.5 * jnp.sum(jnp.square(alpha)) - self.solver.normalization()
        )
        return jnp.where(jnp.isfinite(loglike), loglike, -jnp.inf)

    @partial(jax.jit, static_argnums=0)
    def _get_alpha(self, y: JAXArray) -> JAXArray:
        return self.solver.solve_triangular(y - self.loc)

    @partial(jax.jit, static_argnums=(0, 3))
    def _condition(
        self,
        y: JAXArray,
        X_test: Optional[JAXArray],
        include_mean: bool,
        kernel: Optional[kernels.Kernel] = None,
    ) -> Tuple[JAXArray, JAXArray, JAXArray]:
        alpha = self._get_alpha(y)
        log_prob = self._compute_log_prob(alpha)

        # Below, we actually want alpha = K^-1 y instead of alpha = L^-1 y
        alpha = self.solver.solve_triangular(alpha, transpose=True)

        if X_test is None:
            X_test = self.X

            # In this common case (where we're predicting the GP at the data
            # points, using the original kernel), the mean is especially fast to
            # compute; so let's use that calculation here.
            if kernel is None:
                delta = self.noise @ alpha
                mean_value = y - delta
                if not include_mean:
                    mean_value -= self.loc

            else:
                mean_value = kernel.matmul(self.X, y=alpha)
                if include_mean:
                    mean_value += self.loc

        else:
            if kernel is None:
                kernel = self.kernel

            mean_value = kernel.matmul(X_test, self.X, alpha)
            if include_mean:
                mean_value += jax.vmap(self.mean_function)(X_test)

        return alpha, log_prob, mean_value


class ConditionResult(NamedTuple):
    """The result of conditioning a :class:`GaussianProcess` on data

    This has two entries, ``log_probability`` and ``gp``, that are described
    below.
    """

    log_probability: JAXArray
    """The log probability of the conditioned model

    In other words, this is the marginal likelihood for the kernel parameters,
    given the observed data, or the multivariate normal log probability
    evaluated at the given data.
    """

    gp: GaussianProcess
    """A :class:`GaussianProcess` describing the conditional distribution

    This will have a mean and covariance conditioned on the observed data, but
    it is otherwise a fully functional GP that can sample from or condition
    further (although that's probably not going to be very efficient).
    """


def _default_diag(reference: JAXArray) -> JAXArray:
    """Default to adding some amount of jitter to the diagonal, just in case,
    we use sqrt(eps) for the dtype of the mean function because that seems to
    give sensible results in general.
    """
    return jnp.sqrt(jnp.finfo(reference).eps)



# -*- coding: utf-8 -*-
"""
The algorithms implemented in this subpackage are are mostly based on `Eidelman
& Gohberg (1999) <https://link.springer.com/article/10.1007%2FBF01300581>`_ and
`Foreman-Mackey et al. (2017) <https://arxiv.org/abs/1703.09710>`_.
"""

from __future__ import annotations

__all__ = [
    "DiagQSM",
    "StrictLowerTriQSM",
    "StrictUpperTriQSM",
    "LowerTriQSM",
    "UpperTriQSM",
    "SquareQSM",
    "SymmQSM",
]

from abc import ABCMeta, abstractmethod
from functools import wraps
from typing import Any, Callable, Tuple

import jax
import jax.numpy as jnp
import numpy as np
from jax.scipy.linalg import block_diag

from tinygp.helpers import JAXArray, dataclass


def handle_matvec_shapes(
    func: Callable[[Any, JAXArray], JAXArray]
) -> Callable[[Any, JAXArray], JAXArray]:
    @wraps(func)
    def wrapped(self: Any, x: JAXArray) -> JAXArray:
        output_shape = x.shape
        result = func(self, jnp.reshape(x, (output_shape[0], -1)))
        return jnp.reshape(result, output_shape)

    return wrapped


class QSM(metaclass=ABCMeta):
    """The base class for all square quasiseparable matrices

    This class has blanket implementations of the standard operations that are
    implemented for all QSMs, like addtion, subtraction, multiplication, and
    matrix multiplication.
    """

    # Must be higher than jax's
    __array_priority__ = 2000

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        # Stub for mypy
        raise NotImplementedError

    @abstractmethod
    def transpose(self) -> Any:
        """The matrix transpose as a QSM"""
        raise NotImplementedError

    @abstractmethod
    def matmul(self, x: JAXArray) -> JAXArray:
        """The dot product of this matrix with a dense vector or matrix

        Args:
            x (n, ...): A matrix or vector with leading dimension matching this
                matrix.
        """
        raise NotImplementedError

    @abstractmethod
    def scale(self, other: JAXArray) -> "QSM":
        """The multiplication of this matrix times a scalar, as a QSM"""
        raise NotImplementedError

    @property
    def T(self) -> Any:
        return self.transpose()

    def to_dense(self) -> JAXArray:
        """Render this representation to a dense matrix

        This implementation is not optimized and should really only ever be used
        for testing purposes.
        """
        return self.matmul(jnp.eye(self.shape[0]))

    @property
    def shape(self) -> Tuple[int, int]:
        """The shape of the matrix"""
        n = self.diag.shape[0]  # type: ignore
        return (n, n)

    def __iter__(self):  # type: ignore
        return self.iter_elems()  # type: ignore

    @jax.jit
    def __sub__(self, other: Any) -> Any:
        return self.__add__(-other)

    @jax.jit
    def __add__(self, other: Any) -> Any:
        from tinygp.solvers.quasisep.ops import elementwise_add

        return elementwise_add(self, other)

    @jax.jit
    def __mul__(self, other: Any) -> Any:
        if isinstance(other, QSM):
            from tinygp.solvers.quasisep.ops import elementwise_mul

            return elementwise_mul(self, other)
        else:
            assert jnp.ndim(other) <= 1
            return self.scale(other)

    @jax.jit
    def __rmul__(self, other: Any) -> Any:
        assert not isinstance(other, QSM)
        assert jnp.ndim(other) <= 1
        return self.scale(other)

    @jax.jit
    def __matmul__(self, other: Any) -> Any:
        if isinstance(other, QSM):
            from tinygp.solvers.quasisep.ops import qsm_mul

            return qsm_mul(self, other)
        else:
            return self.matmul(other)

    @jax.jit
    def __rmatmul__(self, other: Any) -> Any:
        assert not isinstance(other, QSM)
        return (self.transpose() @ other.transpose()).transpose()


@dataclass
class DiagQSM(QSM):
    """A diagonal quasiseparable matrix

    Args:
        d (n,): The diagonal entries of the matrix as a 1-D array.
    """

    d: JAXArray

    @property
    def shape(self) -> Tuple[int, int]:
        n = self.d.shape[0]
        return (n, n)

    def transpose(self) -> "DiagQSM":
        return self

    @handle_matvec_shapes
    def matmul(self, x: JAXArray) -> JAXArray:
        return self.d[:, None] * x

    def scale(self, other: JAXArray) -> "DiagQSM":
        return DiagQSM(d=self.d * other)

    def self_add(self, other: "DiagQSM") -> "DiagQSM":
        """The sum of two :class:`DiagQSM` matrices"""
        return DiagQSM(d=self.d + other.d)

    def self_mul(self, other: "DiagQSM") -> "DiagQSM":
        """The elementwise product of two :class:`DiagQSM` matrices"""
        return DiagQSM(d=self.d * other.d)

    def __neg__(self) -> "DiagQSM":
        return DiagQSM(d=-self.d)


@dataclass
class StrictLowerTriQSM(QSM):
    """A strictly lower triangular order ``m`` quasiseparable matrix

    Args:
        p (n, m): The left quasiseparable elements.
        q (n, m): The right quasiseparable elements.
        a (n, m, m): The transition matrices.
    """

    p: JAXArray
    q: JAXArray
    a: JAXArray

    @property
    def shape(self) -> Tuple[int, int]:
        n = self.p.shape[0]
        return (n, n)

    def transpose(self) -> "StrictUpperTriQSM":
        return StrictUpperTriQSM(p=self.p, q=self.q, a=self.a)

    @jax.jit
    @handle_matvec_shapes
    def matmul(self, x: JAXArray) -> JAXArray:
        def impl(f, data):  # type: ignore
            q, a, x = data
            return a @ f + jnp.outer(q, x), f

        init = jnp.zeros_like(jnp.outer(self.q[0], x[0]))
        _, f = jax.lax.scan(impl, init, (self.q, self.a, x))
        return jax.vmap(jnp.dot)(self.p, f)

    def scale(self, other: JAXArray) -> "StrictLowerTriQSM":
        return StrictLowerTriQSM(p=self.p * other, q=self.q, a=self.a)

    def self_add(self, other: "StrictLowerTriQSM") -> "StrictLowerTriQSM":
        """The sum of two :class:`StrictLowerTriQSM` matrices"""

        @jax.vmap
        def impl(
            self: StrictLowerTriQSM, other: StrictLowerTriQSM
        ) -> StrictLowerTriQSM:
            p1, q1, a1 = self
            p2, q2, a2 = other
            return StrictLowerTriQSM(
                p=jnp.concatenate((p1, p2)),
                q=jnp.concatenate((q1, q2)),
                a=block_diag(a1, a2),
            )

        return impl(self, other)

    def self_mul(self, other: "StrictLowerTriQSM") -> "StrictLowerTriQSM":
        """The elementwise product of two :class:`StrictLowerTriQSM` matrices"""
        i, j = np.meshgrid(
            np.arange(self.p.shape[1]), np.arange(other.p.shape[1])
        )
        i = i.flatten()
        j = j.flatten()
        return StrictLowerTriQSM(
            p=self.p[:, i] * other.p[:, j],
            q=self.q[:, i] * other.q[:, j],
            a=self.a[:, i[:, None], i[None, :]]
            * other.a[:, j[:, None], j[None, :]],
        )

    def __neg__(self) -> "StrictLowerTriQSM":
        return StrictLowerTriQSM(p=-self.p, q=self.q, a=self.a)


@dataclass
class StrictUpperTriQSM(QSM):
    """A strictly upper triangular order ``m`` quasiseparable matrix

    The notation here is somewhat different from that in `Eidelman & Gohberg
    (1999) <https://link.springer.com/article/10.1007%2FBF01300581>`_, because
    we wanted to map ``StrictLowerTriQSM.transpose() -> StrictUpperTriQSM``
    while retaining the same names for each component. Therefore, our ``p`` is
    their ``h``, and our ``a`` is their ``b.T``.

    Args:
        p (n, m): The right quasiseparable elements.
        q (n, m): The left quasiseparable elements.
        a (n, m, m): The transition matrices.
    """

    p: JAXArray
    q: JAXArray
    a: JAXArray

    @property
    def shape(self) -> Tuple[int, int]:
        n = self.p.shape[0]
        return (n, n)

    def transpose(self) -> "StrictLowerTriQSM":
        return StrictLowerTriQSM(p=self.p, q=self.q, a=self.a)

    @jax.jit
    @handle_matvec_shapes
    def matmul(self, x: JAXArray) -> JAXArray:
        def impl(f, data):  # type: ignore
            p, a, x = data
            return a.T @ f + jnp.outer(p, x), f

        init = jnp.zeros_like(jnp.outer(self.p[-1], x[-1]))
        _, f = jax.lax.scan(impl, init, (self.p, self.a, x), reverse=True)
        return jax.vmap(jnp.dot)(self.q, f)

    def scale(self, other: JAXArray) -> "StrictUpperTriQSM":
        return StrictUpperTriQSM(p=self.p, q=self.q * other, a=self.a)

    def self_add(self, other: "StrictUpperTriQSM") -> "StrictUpperTriQSM":
        """The sum of two :class:`StrictUpperTriQSM` matrices"""
        return self.transpose().self_add(other.transpose()).transpose()

    def self_mul(self, other: "StrictUpperTriQSM") -> "StrictUpperTriQSM":
        """The elementwise product of two :class:`StrictUpperTriQSM` matrices"""
        return self.transpose().self_mul(other.transpose()).transpose()

    def __neg__(self) -> "StrictUpperTriQSM":
        return StrictUpperTriQSM(p=-self.p, q=self.q, a=self.a)


@dataclass
class LowerTriQSM(QSM):
    """A lower triangular quasiseparable matrix

    Args:
        diag: The diagonal elements.
        lower: The strictly lower triangular elements.
    """

    diag: DiagQSM
    lower: StrictLowerTriQSM

    def transpose(self) -> "UpperTriQSM":
        return UpperTriQSM(diag=self.diag, upper=self.lower.transpose())

    @handle_matvec_shapes
    def matmul(self, x: JAXArray) -> JAXArray:
        return self.diag.matmul(x) + self.lower.matmul(x)

    def scale(self, other: JAXArray) -> "LowerTriQSM":
        return LowerTriQSM(
            diag=self.diag.scale(other), lower=self.lower.scale(other)
        )

    def inv(self) -> "LowerTriQSM":
        (d,) = self.diag
        p, q, a = self.lower
        g = 1 / d
        u = -g[:, None] * p
        v = g[:, None] * q
        b = a - jax.vmap(jnp.outer)(v, p)
        return LowerTriQSM(
            diag=DiagQSM(g), lower=StrictLowerTriQSM(p=u, q=v, a=b)
        )

    @jax.jit
    @handle_matvec_shapes
    def solve(self, y: JAXArray) -> JAXArray:
        """Solve a linear system with this matrix

        If this matrix is called ``L``, this solves ``L @ x = y`` for ``x``
        given ``y``, using forward substitution.

        Args:
            y (n, ...): A matrix or vector with leading dimension matching this
                matrix.
        """

        def impl(fn, data):  # type: ignore
            ((cn,), (pn, wn, an)), yn = data
            xn = (yn - pn @ fn) / cn
            return an @ fn + jnp.outer(wn, xn), xn

        init = jnp.zeros_like(jnp.outer(self.lower.q[0], y[0]))
        _, x = jax.lax.scan(impl, init, (self, y))
        return x

    def __neg__(self) -> "LowerTriQSM":
        return LowerTriQSM(diag=-self.diag, lower=-self.lower)


@dataclass
class UpperTriQSM(QSM):
    """A upper triangular quasiseparable matrix

    Args:
        diag: The diagonal elements.
        upper: The strictly upper triangular elements.
    """

    diag: DiagQSM
    upper: StrictUpperTriQSM

    def transpose(self) -> "LowerTriQSM":
        return LowerTriQSM(diag=self.diag, lower=self.upper.transpose())

    @handle_matvec_shapes
    def matmul(self, x: JAXArray) -> JAXArray:
        return self.diag.matmul(x) + self.upper.matmul(x)

    def scale(self, other: JAXArray) -> "UpperTriQSM":
        return UpperTriQSM(
            diag=self.diag.scale(other), upper=self.upper.scale(other)
        )

    def inv(self) -> "UpperTriQSM":
        return self.transpose().inv().transpose()

    @jax.jit
    @handle_matvec_shapes
    def solve(self, y: JAXArray) -> JAXArray:
        """Solve a linear system with this matrix

        If this matrix is called ``U``, this solves ``U @ x = y`` for ``x``
        given ``y``, using backward substitution.

        Args:
            y (n, ...): A matrix or vector with leading dimension matching this
                matrix.
        """

        def impl(fn, data):  # type: ignore
            ((cn,), (pn, wn, an)), yn = data
            xn = (yn - wn @ fn) / cn
            return an.T @ fn + jnp.outer(pn, xn), xn

        init = jnp.zeros_like(jnp.outer(self.upper.p[-1], y[-1]))
        _, x = jax.lax.scan(impl, init, (self, y), reverse=True)
        return x

    def __neg__(self) -> "UpperTriQSM":
        return UpperTriQSM(diag=-self.diag, upper=-self.upper)


@dataclass
class SquareQSM(QSM):
    """A general square order ``(m1, m2)`` quasiseparable matrix

    Args:
        diag: The diagonal elements.
        lower: The strictly lower triangular elements with order ``m1``.
        upper: The strictly upper triangular elements with order ``m2``.
    """

    diag: DiagQSM
    lower: StrictLowerTriQSM
    upper: StrictUpperTriQSM

    def transpose(self) -> "SquareQSM":
        return SquareQSM(
            diag=self.diag,
            lower=self.upper.transpose(),
            upper=self.lower.transpose(),
        )

    @handle_matvec_shapes
    def matmul(self, x: JAXArray) -> JAXArray:
        return (
            self.diag.matmul(x) + self.lower.matmul(x) + self.upper.matmul(x)
        )

    def scale(self, other: JAXArray) -> "SquareQSM":
        return SquareQSM(
            diag=self.diag.scale(other),
            lower=self.lower.scale(other),
            upper=self.upper.scale(other),
        )

    def gram(self) -> "SymmQSM":
        """The inner product of this matrix with itself

        If this matrix is called ``A``, the Gram matrix is ``A.T @ A``, and
        that's what this method computes. The result is a :class:`SymmQSM`.
        """
        # We know that this must result in symmetric matrix, but that won't be
        # enforced; we make it so! It might be possible to make this more
        # efficient, but perhaps jax is clever enough?
        M = self.transpose() @ self
        return SymmQSM(diag=M.diag, lower=M.lower)

    @jax.jit
    def inv(self) -> "SquareQSM":
        """The inverse of this matrix"""
        (d,) = self.diag
        p, q, a = self.lower
        h, g, b = self.upper

        def forward(carry, data):  # type: ignore
            f = carry
            dk, pk, qk, ak, gk, hk, bk = data
            fhk = f @ hk
            fbk = f @ bk.T
            left = qk - ak @ fhk
            right = gk - pk @ fbk
            igk = 1 / (dk - pk @ fhk)
            sk = igk * left
            ellk = ak - jnp.outer(sk, pk)
            vk = igk * right
            delk = bk - jnp.outer(vk, hk)
            fk = ak @ fbk + igk * jnp.outer(left, right)
            return fk, (igk, sk, ellk, vk, delk)

        init = jnp.zeros_like(jnp.outer(q[0], g[0]))
        ig, s, ell, v, del_ = jax.lax.scan(
            forward, init, (d, p, q, a, g, h, b)
        )[1]

        def backward(carry, data):  # type: ignore
            z = carry
            igk, pk, ak, hk, bk, sk, vk = data
            zsk = z @ sk
            zak = z @ ak
            lk = igk + vk @ zsk
            tk = vk @ zak - lk * pk
            uk = bk.T @ zsk - lk * hk
            zk = bk.T @ zak - jnp.outer(uk + lk * hk, pk) - jnp.outer(hk, tk)
            return zk, (lk, tk, uk)

        init = jnp.zeros_like(jnp.outer(h[-1], p[-1]))
        args = (ig, p, a, h, b, s, v)
        lam, t, u = jax.lax.scan(backward, init, args, reverse=True)[1]
        return SquareQSM(
            diag=DiagQSM(d=lam),
            lower=StrictLowerTriQSM(p=t, q=s, a=ell),
            upper=StrictUpperTriQSM(p=u, q=v, a=del_),
        )

    def __neg__(self) -> "SquareQSM":
        return SquareQSM(diag=-self.diag, lower=-self.lower, upper=-self.upper)


@dataclass
class SymmQSM(QSM):
    """A symmetric order ``m`` quasiseparable matrix

    Args:
        diag: The diagonal elements.
        lower: The strictly lower triangular elements with order ``m``.
    """

    diag: DiagQSM
    lower: StrictLowerTriQSM

    def transpose(self) -> "SymmQSM":
        return self

    @handle_matvec_shapes
    def matmul(self, x: JAXArray) -> JAXArray:
        return (
            self.diag.matmul(x)
            + self.lower.matmul(x)
            + self.lower.transpose().matmul(x)
        )

    def scale(self, other: JAXArray) -> "SymmQSM":
        return SymmQSM(
            diag=self.diag.scale(other), lower=self.lower.scale(other)
        )

    @jax.jit
    def inv(self) -> "SymmQSM":
        """The inverse of this matrix"""
        (d,) = self.diag
        p, q, a = self.lower

        def forward(carry, data):  # type: ignore
            f = carry
            dk, pk, qk, ak = data
            fpk = f @ pk
            left = qk - ak @ fpk
            igk = 1 / (dk - pk @ fpk)
            sk = igk * left
            ellk = ak - jnp.outer(sk, pk)
            fk = ak @ f @ ak.T + igk * jnp.outer(left, left.T)
            return fk, (igk, sk, ellk)

        init = jnp.zeros_like(jnp.outer(q[0], q[0]))
        ig, s, ell = jax.lax.scan(forward, init, (d, p, q, a))[1]

        def backward(carry, data):  # type: ignore
            z = carry
            igk, pk, ak, sk = data
            zak = z @ ak
            skzak = sk @ zak
            lk = igk + sk @ z @ sk
            tk = skzak - lk * pk
            zk = ak.T @ zak - jnp.outer(skzak, pk) - jnp.outer(pk, tk)
            return zk, (lk, tk)

        init = jnp.zeros_like(jnp.outer(p[-1], p[-1]))
        lam, t = jax.lax.scan(backward, init, (ig, p, a, s), reverse=True)[1]
        return SymmQSM(
            diag=DiagQSM(d=lam), lower=StrictLowerTriQSM(p=t, q=s, a=ell)
        )

    @jax.jit
    def cholesky(self) -> LowerTriQSM:
        """The Cholesky decomposition of this matrix

        If this matrix is called ``A``, this method returns the
        :class:`LowerTriQSM` ``L`` such that ``L @ L.T = A``.
        """
        (d,) = self.diag
        p, q, a = self.lower

        def impl(carry, data):  # type: ignore
            fp = carry
            dk, pk, qk, ak = data
            ck = jnp.sqrt(dk - pk @ fp @ pk)
            tmp = fp @ ak.T
            wk = (qk - pk @ tmp) / ck
            fk = ak @ tmp + jnp.outer(wk, wk)
            return fk, (ck, wk)

        init = jnp.zeros_like(jnp.outer(q[0], q[0]))
        _, (c, w) = jax.lax.scan(impl, init, (d, p, q, a))
        return LowerTriQSM(
            diag=DiagQSM(c), lower=StrictLowerTriQSM(p=p, q=w, a=a)
        )

    def __neg__(self) -> "SymmQSM":
        return SymmQSM(diag=-self.diag, lower=-self.lower)



# -*- coding: utf-8 -*-
# mypy: ignore-errors

import jax
import jax.numpy as jnp
import numpy as np
import pytest

from tinygp import GaussianProcess, kernels


@pytest.fixture
def random():
    return np.random.default_rng(1058390)


@pytest.fixture
def data(random):
    X = random.uniform(-3, 3, (50, 5))
    y = random.normal(len(X))
    return X, y


def test_sample(data):
    X, _ = data

    with jax.experimental.enable_x64(True):
        gp = GaussianProcess(
            kernels.Matern32(1.5), X, diag=0.01, mean=lambda x: jnp.sum(x)
        )
        y = gp.sample(jax.random.PRNGKey(543))
        assert y.shape == (len(X),)

        y = gp.sample(jax.random.PRNGKey(543), shape=(7, 3))
        assert y.shape == (7, 3, len(X))

        y = gp.sample(jax.random.PRNGKey(543), shape=(100_000,))
        assert y.shape == (100_000, len(X))
        np.testing.assert_allclose(
            jnp.mean(y, axis=0), jnp.sum(X, axis=1), atol=0.015
        )
        np.testing.assert_allclose(
            jnp.cov(y, rowvar=0), gp.covariance, atol=0.015
        )


def test_means(data):
    X, y = data

    gp1 = GaussianProcess(
        kernels.Matern32(1.5), X, diag=0.01, mean=lambda x: 0.0
    )
    gp2 = GaussianProcess(kernels.Matern32(1.5), X, diag=0.01, mean=0.0)
    gp3 = GaussianProcess(kernels.Matern32(1.5), X, diag=0.01)

    np.testing.assert_allclose(gp1.mean, gp2.mean)
    np.testing.assert_allclose(gp1.mean, gp3.mean)
    np.testing.assert_allclose(gp1.log_probability(y), gp2.log_probability(y))
    np.testing.assert_allclose(gp1.log_probability(y), gp3.log_probability(y))


@pytest.mark.parametrize("tree", [True, False])
def test_condition_shape_error(data, tree):
    if tree:

        class CustomDistance(kernels.Distance):
            def distance(self, X1, X2):
                return kernels.L2Distance().distance(X1["x"], X2["x"])

        distance = CustomDistance()
    else:
        distance = kernels.L2Distance()

    X, y = data
    kernel = kernels.ExpSquared(distance=distance)
    gp = GaussianProcess(kernel, {"x": X} if tree else X, diag=0.1)
    gp.condition(y, {"x": X[0][None]} if tree else X[0][None])

    with pytest.raises(ValueError):
        gp.condition(y, X[0])

    if tree:
        with pytest.raises(ValueError):
            gp.condition(y, {"x": X[0]})



# -*- coding: utf-8 -*-
"""
``tinygp`` is an extremely lightweight library for building Gaussian Process
models in Python, built on top of `jax <https://github.com/google/jax>`_. The
primary way that you will use to interact with ``tinygp`` is by constructing
"kernel" functions using the building blocks provided in the ``kernels``
subpackage (see :ref:`api-kernels`), and then passing that to a
:class:`GaussianProcess` object to do all the computations. Check out the
:ref:`tutorials` for a more complete introduction.
"""

__all__ = [
    "__version__",
    "kernels",
    "noise",
    "solvers",
    "transforms",
    "GaussianProcess",
]

from tinygp import kernels, noise, solvers, transforms
from tinygp.gp import GaussianProcess
from tinygp.tinygp_version import __version__

__author__ = "Dan Foreman-Mackey"
__email__ = "foreman.mackey@gmail.com"
__uri__ = "https://github.com/dfm/tinygp"
__license__ = "MIT"
__description__ = "The tiniest of Gaussian Process libraries"
__copyright__ = "2021, 2022 Simons Foundation, Inc"
__contributors__ = "https://github.com/dfm/tinygp/graphs/contributors"
__bibtex__ = __citation__ = """TBD"""



# -*- coding: utf-8 -*-

from __future__ import annotations

__all__ = ["Solver"]

from abc import ABCMeta, abstractmethod
from typing import Any, Optional

from tinygp.helpers import JAXArray
from tinygp.kernels.base import Kernel
from tinygp.noise import Noise


class Solver(metaclass=ABCMeta):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    @classmethod
    def init(
        cls,
        kernel: Kernel,
        X: JAXArray,
        noise: Noise,
        *,
        covariance: Optional[Any] = None,
    ) -> "Solver":
        raise NotImplementedError

    @abstractmethod
    def variance(self) -> JAXArray:
        """The diagonal of the covariance matrix"""
        raise NotImplementedError

    @abstractmethod
    def covariance(self) -> JAXArray:
        """The evaluated covariance matrix"""
        raise NotImplementedError

    @abstractmethod
    def normalization(self) -> JAXArray:
        """The multivariate normal normalization constant

        This should be ``(log_det + n*log(2*pi))/2``, where ``n`` is the size of
        the covariance matrix, and ``log_det`` is the log determinant of the
        matrix.
        """
        raise NotImplementedError

    @abstractmethod
    def solve_triangular(
        self, y: JAXArray, *, transpose: bool = False
    ) -> JAXArray:
        """Solve the lower triangular linear system defined by this solver

        If the covariance matrix is ``K = L @ L.T`` for some lower triangular
        matrix ``L``, this method solves ``L @ x = y`` for some ``y``. If the
        ``transpose`` parameter is ``True``, this instead solves ``L.T @ x =
        y``.
        """
        raise NotImplementedError

    @abstractmethod
    def dot_triangular(self, y: JAXArray) -> JAXArray:
        """Compute a matrix product with the lower triangular linear system

        If the covariance matrix is ``K = L @ L.T`` for some lower triangular
        matrix ``L``, this method returns ``L @ y`` for some ``y``.
        """
        raise NotImplementedError

    @abstractmethod
    def condition(
        self, kernel: Kernel, X_test: Optional[JAXArray], noise: Noise
    ) -> Any:
        raise NotImplementedError
